# Licensed under a 3-clause BSD style license - see LICENSE.rst

import operator

import numpy as np
import pytest

from astropy import units as u
from astropy.coordinates import (
    CartesianDifferential,
    CartesianRepresentation,
    CylindricalDifferential,
    CylindricalRepresentation,
    Latitude,
    Longitude,
    PhysicsSphericalDifferential,
    PhysicsSphericalRepresentation,
    RadialDifferential,
    RadialRepresentation,
    SphericalCosLatDifferential,
    SphericalDifferential,
    SphericalRepresentation,
    UnitSphericalCosLatDifferential,
    UnitSphericalDifferential,
    UnitSphericalRepresentation,
    angular_separation,
)
from astropy.coordinates.representation import DIFFERENTIAL_CLASSES
from astropy.tests.helper import assert_quantity_allclose, quantity_allclose


def assert_representation_allclose(actual, desired, rtol=1.0e-7, atol=None, **kwargs):
    __tracebackhide__ = True
    actual_xyz = actual.to_cartesian().get_xyz(xyz_axis=-1)
    desired_xyz = desired.to_cartesian().get_xyz(xyz_axis=-1)
    actual_xyz, desired_xyz = np.broadcast_arrays(actual_xyz, desired_xyz, subok=True)
    assert_quantity_allclose(actual_xyz, desired_xyz, rtol, atol, **kwargs)


def assert_differential_allclose(actual, desired, rtol=1.0e-7, **kwargs):
    assert actual.components == desired.components
    for component in actual.components:
        actual_c = getattr(actual, component)
        atol = 1.0e-10 * actual_c.unit
        assert_quantity_allclose(
            actual_c, getattr(desired, component), rtol, atol, **kwargs
        )


def representation_equal(first, second):
    return np.all(
        getattr(first, comp) == getattr(second, comp) for comp in first.components
    )


class TestArithmetic:
    def setup_method(self):
        # Choose some specific coordinates, for which ``sum`` and ``dot``
        # works out nicely.
        self.lon = Longitude(np.arange(0, 12.1, 2), u.hourangle)
        self.lat = Latitude(np.arange(-90, 91, 30), u.deg)
        self.distance = [5.0, 12.0, 4.0, 2.0, 4.0, 12.0, 5.0] * u.kpc
        self.spherical = SphericalRepresentation(self.lon, self.lat, self.distance)
        self.unit_spherical = self.spherical.represent_as(UnitSphericalRepresentation)
        self.cartesian = self.spherical.to_cartesian()

    def test_norm_spherical(self):
        norm_s = self.spherical.norm()
        assert isinstance(norm_s, u.Quantity)
        # Just to be sure, test against getting object arrays.
        assert norm_s.dtype.kind == "f"
        assert np.all(norm_s == self.distance)

    @pytest.mark.parametrize(
        "representation",
        (
            PhysicsSphericalRepresentation,
            CartesianRepresentation,
            CylindricalRepresentation,
        ),
    )
    def test_norm(self, representation):
        in_rep = self.spherical.represent_as(representation)
        norm_rep = in_rep.norm()
        assert isinstance(norm_rep, u.Quantity)
        assert_quantity_allclose(norm_rep, self.distance)

    def test_norm_unitspherical(self):
        norm_rep = self.unit_spherical.norm()
        assert norm_rep.unit == u.dimensionless_unscaled
        assert np.all(norm_rep == 1.0 * u.dimensionless_unscaled)

    @pytest.mark.parametrize(
        "representation",
        (
            SphericalRepresentation,
            PhysicsSphericalRepresentation,
            CartesianRepresentation,
            CylindricalRepresentation,
            UnitSphericalRepresentation,
        ),
    )
    def test_neg_pos(self, representation):
        in_rep = self.cartesian.represent_as(representation)
        pos_rep = +in_rep
        assert type(pos_rep) is type(in_rep)
        assert pos_rep is not in_rep
        assert np.all(representation_equal(pos_rep, in_rep))
        neg_rep = -in_rep
        assert type(neg_rep) is type(in_rep)
        assert np.all(neg_rep.norm() == in_rep.norm())
        in_rep_xyz = in_rep.to_cartesian().xyz
        assert_quantity_allclose(
            neg_rep.to_cartesian().xyz, -in_rep_xyz, atol=1.0e-10 * in_rep_xyz.unit
        )

    def test_mul_div_spherical(self):
        s0 = self.spherical / (1.0 * u.Myr)
        assert isinstance(s0, SphericalRepresentation)
        assert s0.distance.dtype.kind == "f"
        assert np.all(s0.lon == self.spherical.lon)
        assert np.all(s0.lat == self.spherical.lat)
        assert np.all(s0.distance == self.distance / (1.0 * u.Myr))
        s1 = (1.0 / u.Myr) * self.spherical
        assert isinstance(s1, SphericalRepresentation)
        assert np.all(representation_equal(s1, s0))
        s2 = self.spherical * np.array([[1.0], [2.0]])
        assert isinstance(s2, SphericalRepresentation)
        assert s2.shape == (2, self.spherical.shape[0])
        assert np.all(s2.lon == self.spherical.lon)
        assert np.all(s2.lat == self.spherical.lat)
        assert np.all(s2.distance == self.spherical.distance * np.array([[1.0], [2.0]]))
        s3 = np.array([[1.0], [2.0]]) * self.spherical
        assert isinstance(s3, SphericalRepresentation)
        assert np.all(representation_equal(s3, s2))
        s4 = -self.spherical
        assert isinstance(s4, SphericalRepresentation)
        assert quantity_allclose(
            s4.to_cartesian().xyz,
            -self.spherical.to_cartesian().xyz,
            atol=1e-15 * self.spherical.distance.unit,
        )
        assert np.all(s4.distance == self.spherical.distance)
        s5 = +self.spherical
        assert s5 is not self.spherical
        assert np.all(representation_equal(s5, self.spherical))

    @pytest.mark.parametrize(
        "representation",
        (
            PhysicsSphericalRepresentation,
            CartesianRepresentation,
            CylindricalRepresentation,
        ),
    )
    def test_mul_div(self, representation):
        in_rep = self.spherical.represent_as(representation)
        r1 = in_rep / (1.0 * u.Myr)
        assert isinstance(r1, representation)
        for component in in_rep.components:
            in_rep_comp = getattr(in_rep, component)
            r1_comp = getattr(r1, component)
            if in_rep_comp.unit == self.distance.unit:
                assert np.all(r1_comp == in_rep_comp / (1.0 * u.Myr))
            else:
                assert np.all(r1_comp == in_rep_comp)

        r2 = np.array([[1.0], [2.0]]) * in_rep
        assert isinstance(r2, representation)
        assert r2.shape == (2, in_rep.shape[0])
        assert_quantity_allclose(r2.norm(), self.distance * np.array([[1.0], [2.0]]))
        r3 = -in_rep
        assert_representation_allclose(
            r3.to_cartesian(), (in_rep * -1.0).to_cartesian(), atol=1e-5 * u.pc
        )
        with pytest.raises(TypeError):
            in_rep * in_rep
        with pytest.raises(TypeError):
            {} * in_rep

    def test_mul_div_unit_spherical(self):
        s1 = self.unit_spherical * self.distance
        assert isinstance(s1, SphericalRepresentation)
        assert np.all(s1.lon == self.unit_spherical.lon)
        assert np.all(s1.lat == self.unit_spherical.lat)
        assert np.all(s1.distance == self.spherical.distance)
        s2 = self.unit_spherical / u.s
        assert isinstance(s2, SphericalRepresentation)
        assert np.all(s2.lon == self.unit_spherical.lon)
        assert np.all(s2.lat == self.unit_spherical.lat)
        assert np.all(s2.distance == 1.0 / u.s)
        u3 = -self.unit_spherical
        assert isinstance(u3, UnitSphericalRepresentation)
        assert_quantity_allclose(u3.lon, self.unit_spherical.lon + 180.0 * u.deg)
        assert np.all(u3.lat == -self.unit_spherical.lat)
        assert_quantity_allclose(
            u3.to_cartesian().xyz,
            -self.unit_spherical.to_cartesian().xyz,
            atol=1.0e-10 * u.dimensionless_unscaled,
        )
        u4 = +self.unit_spherical
        assert isinstance(u4, UnitSphericalRepresentation)
        assert u4 is not self.unit_spherical
        assert np.all(representation_equal(u4, self.unit_spherical))

    def test_add_sub_cartesian(self):
        c1 = self.cartesian + self.cartesian
        assert isinstance(c1, CartesianRepresentation)
        assert c1.x.dtype.kind == "f"
        assert np.all(representation_equal(c1, 2.0 * self.cartesian))
        with pytest.raises(TypeError):
            self.cartesian + 10.0 * u.m
        with pytest.raises(u.UnitsError):
            self.cartesian + (self.cartesian / u.s)
        c2 = self.cartesian - self.cartesian
        assert isinstance(c2, CartesianRepresentation)
        assert np.all(
            representation_equal(
                c2, CartesianRepresentation(0.0 * u.m, 0.0 * u.m, 0.0 * u.m)
            )
        )
        c3 = self.cartesian - self.cartesian / 2.0
        assert isinstance(c3, CartesianRepresentation)
        assert np.all(representation_equal(c3, self.cartesian / 2.0))

    @pytest.mark.parametrize(
        "representation",
        (
            PhysicsSphericalRepresentation,
            SphericalRepresentation,
            CylindricalRepresentation,
        ),
    )
    def test_add_sub(self, representation):
        in_rep = self.cartesian.represent_as(representation)
        r1 = in_rep + in_rep
        assert isinstance(r1, representation)
        expected = 2.0 * in_rep
        for component in in_rep.components:
            assert_quantity_allclose(
                getattr(r1, component), getattr(expected, component)
            )
        with pytest.raises(TypeError):
            10.0 * u.m + in_rep
        with pytest.raises(u.UnitsError):
            in_rep + (in_rep / u.s)
        r2 = in_rep - in_rep
        assert isinstance(r2, representation)
        assert_representation_allclose(
            r2.to_cartesian(),
            CartesianRepresentation(0.0 * u.m, 0.0 * u.m, 0.0 * u.m),
            atol=1e-15 * u.kpc,
        )
        r3 = in_rep - in_rep / 2.0
        assert isinstance(r3, representation)
        expected = in_rep / 2.0
        assert_representation_allclose(r3, expected)

    def test_add_sub_unit_spherical(self):
        s1 = self.unit_spherical + self.unit_spherical
        assert isinstance(s1, SphericalRepresentation)
        expected = 2.0 * self.unit_spherical
        for component in s1.components:
            assert_quantity_allclose(
                getattr(s1, component), getattr(expected, component)
            )
        with pytest.raises(TypeError):
            10.0 * u.m - self.unit_spherical
        with pytest.raises(u.UnitsError):
            self.unit_spherical + (self.unit_spherical / u.s)
        s2 = self.unit_spherical - self.unit_spherical / 2.0
        assert isinstance(s2, SphericalRepresentation)
        expected = self.unit_spherical / 2.0
        for component in s2.components:
            assert_quantity_allclose(
                getattr(s2, component), getattr(expected, component)
            )

    @pytest.mark.parametrize(
        "representation",
        (
            CartesianRepresentation,
            PhysicsSphericalRepresentation,
            SphericalRepresentation,
            CylindricalRepresentation,
        ),
    )
    def test_sum_mean(self, representation):
        in_rep = self.spherical.represent_as(representation)
        r_sum = in_rep.sum()
        assert isinstance(r_sum, representation)
        expected = SphericalRepresentation(
            90.0 * u.deg, 0.0 * u.deg, 14.0 * u.kpc
        ).represent_as(representation)
        for component in expected.components:
            exp_component = getattr(expected, component)
            assert_quantity_allclose(
                getattr(r_sum, component),
                exp_component,
                atol=1e-10 * exp_component.unit,
            )

        r_mean = in_rep.mean()
        assert isinstance(r_mean, representation)
        expected = expected / len(in_rep)
        for component in expected.components:
            exp_component = getattr(expected, component)
            assert_quantity_allclose(
                getattr(r_mean, component),
                exp_component,
                atol=1e-10 * exp_component.unit,
            )

    def test_sum_mean_unit_spherical(self):
        s_sum = self.unit_spherical.sum()
        assert isinstance(s_sum, SphericalRepresentation)
        expected = SphericalRepresentation(
            90.0 * u.deg, 0.0 * u.deg, 3.0 * u.dimensionless_unscaled
        )
        for component in expected.components:
            exp_component = getattr(expected, component)
            assert_quantity_allclose(
                getattr(s_sum, component),
                exp_component,
                atol=1e-10 * exp_component.unit,
            )

        s_mean = self.unit_spherical.mean()
        assert isinstance(s_mean, SphericalRepresentation)
        expected = expected / len(self.unit_spherical)
        for component in expected.components:
            exp_component = getattr(expected, component)
            assert_quantity_allclose(
                getattr(s_mean, component),
                exp_component,
                atol=1e-10 * exp_component.unit,
            )

    @pytest.mark.parametrize(
        "representation",
        (
            CartesianRepresentation,
            PhysicsSphericalRepresentation,
            SphericalRepresentation,
            CylindricalRepresentation,
        ),
    )
    def test_dot(self, representation):
        in_rep = self.cartesian.represent_as(representation)
        r_dot_r = in_rep.dot(in_rep)
        assert isinstance(r_dot_r, u.Quantity)
        assert r_dot_r.shape == in_rep.shape
        assert_quantity_allclose(np.sqrt(r_dot_r), self.distance)
        r_dot_r_rev = in_rep.dot(in_rep[::-1])
        assert isinstance(r_dot_r_rev, u.Quantity)
        assert r_dot_r_rev.shape == in_rep.shape
        expected = [-25.0, -126.0, 2.0, 4.0, 2.0, -126.0, -25.0] * u.kpc**2
        assert_quantity_allclose(r_dot_r_rev, expected)
        for axis in "xyz":
            project = CartesianRepresentation(
                *(
                    (1.0 if axis == _axis else 0.0) * u.dimensionless_unscaled
                    for _axis in "xyz"
                )
            )
            assert_quantity_allclose(
                in_rep.dot(project), getattr(self.cartesian, axis), atol=1.0 * u.upc
            )
        with pytest.raises(TypeError):
            in_rep.dot(self.cartesian.xyz)

    def test_dot_unit_spherical(self):
        u_dot_u = self.unit_spherical.dot(self.unit_spherical)
        assert isinstance(u_dot_u, u.Quantity)
        assert u_dot_u.shape == self.unit_spherical.shape
        assert_quantity_allclose(u_dot_u, 1.0 * u.dimensionless_unscaled)
        cartesian = self.unit_spherical.to_cartesian()
        for axis in "xyz":
            project = CartesianRepresentation(
                *(
                    (1.0 if axis == _axis else 0.0) * u.dimensionless_unscaled
                    for _axis in "xyz"
                )
            )
            assert_quantity_allclose(
                self.unit_spherical.dot(project), getattr(cartesian, axis), atol=1.0e-10
            )

    @pytest.mark.parametrize(
        "representation",
        (
            CartesianRepresentation,
            PhysicsSphericalRepresentation,
            SphericalRepresentation,
            CylindricalRepresentation,
        ),
    )
    def test_cross(self, representation):
        in_rep = self.cartesian.represent_as(representation)
        r_cross_r = in_rep.cross(in_rep)
        assert isinstance(r_cross_r, representation)
        assert_quantity_allclose(r_cross_r.norm(), 0.0 * u.kpc**2, atol=1.0 * u.mpc**2)
        r_cross_r_rev = in_rep.cross(in_rep[::-1])
        sep = angular_separation(self.lon, self.lat, self.lon[::-1], self.lat[::-1])
        expected = self.distance * self.distance[::-1] * np.sin(sep)
        assert_quantity_allclose(r_cross_r_rev.norm(), expected, atol=1.0 * u.mpc**2)
        unit_vectors = CartesianRepresentation(
            [1.0, 0.0, 0.0] * u.one, [0.0, 1.0, 0.0] * u.one, [0.0, 0.0, 1.0] * u.one
        )[:, np.newaxis]
        r_cross_uv = in_rep.cross(unit_vectors)
        assert r_cross_uv.shape == (3, 7)
        assert_quantity_allclose(
            r_cross_uv.dot(unit_vectors), 0.0 * u.kpc, atol=1.0 * u.upc
        )
        assert_quantity_allclose(
            r_cross_uv.dot(in_rep), 0.0 * u.kpc**2, atol=1.0 * u.mpc**2
        )
        zeros = np.zeros(len(in_rep)) * u.kpc
        expected = CartesianRepresentation(
            u.Quantity((zeros, -self.cartesian.z, self.cartesian.y)),
            u.Quantity((self.cartesian.z, zeros, -self.cartesian.x)),
            u.Quantity((-self.cartesian.y, self.cartesian.x, zeros)),
        )
        # Comparison with spherical is hard since some distances are zero,
        # implying the angles are undefined.
        r_cross_uv_cartesian = r_cross_uv.to_cartesian()
        assert_representation_allclose(r_cross_uv_cartesian, expected, atol=1.0 * u.upc)
        # A final check, with the side benefit of ensuring __truediv__ and norm
        # work on multi-D representations.
        r_cross_uv_by_distance = r_cross_uv / self.distance
        uv_sph = unit_vectors.represent_as(UnitSphericalRepresentation)
        sep = angular_separation(self.lon, self.lat, uv_sph.lon, uv_sph.lat)
        assert_quantity_allclose(r_cross_uv_by_distance.norm(), np.sin(sep), atol=1e-9)

        with pytest.raises(TypeError):
            in_rep.cross(self.cartesian.xyz)

    def test_cross_unit_spherical(self):
        u_cross_u = self.unit_spherical.cross(self.unit_spherical)
        assert isinstance(u_cross_u, SphericalRepresentation)
        assert_quantity_allclose(u_cross_u.norm(), 0.0 * u.one, atol=1.0e-10 * u.one)
        u_cross_u_rev = self.unit_spherical.cross(self.unit_spherical[::-1])
        assert isinstance(u_cross_u_rev, SphericalRepresentation)
        sep = angular_separation(self.lon, self.lat, self.lon[::-1], self.lat[::-1])
        expected = np.sin(sep)
        assert_quantity_allclose(u_cross_u_rev.norm(), expected, atol=1.0e-10 * u.one)


class TestUnitVectorsAndScales:
    @staticmethod
    def check_unit_vectors(e):
        for v in e.values():
            assert type(v) is CartesianRepresentation
            assert_quantity_allclose(v.norm(), 1.0 * u.one)
        return e

    @staticmethod
    def check_scale_factors(sf, rep):
        unit = rep.norm().unit
        for c, f in sf.items():
            assert type(f) is u.Quantity
            assert (f.unit * getattr(rep, c).unit).is_equivalent(unit)

    def test_spherical(self):
        s = SphericalRepresentation(
            lon=[0.0, 6.0, 21.0] * u.hourangle,
            lat=[0.0, -30.0, 85.0] * u.deg,
            distance=[1, 2, 3] * u.kpc,
        )
        e = s.unit_vectors()
        self.check_unit_vectors(e)
        sf = s.scale_factors()
        self.check_scale_factors(sf, s)

        s_lon = s + s.distance * 1e-5 * np.cos(s.lat) * e["lon"]
        assert_quantity_allclose(s_lon.lon, s.lon + 1e-5 * u.rad, atol=1e-10 * u.rad)
        assert_quantity_allclose(s_lon.lat, s.lat, atol=1e-10 * u.rad)
        assert_quantity_allclose(s_lon.distance, s.distance)
        s_lon2 = s + 1e-5 * u.radian * sf["lon"] * e["lon"]
        assert_representation_allclose(s_lon2, s_lon)

        s_lat = s + s.distance * 1e-5 * e["lat"]
        assert_quantity_allclose(s_lat.lon, s.lon)
        assert_quantity_allclose(s_lat.lat, s.lat + 1e-5 * u.rad, atol=1e-10 * u.rad)
        assert_quantity_allclose(s_lon.distance, s.distance)
        s_lat2 = s + 1.0e-5 * u.radian * sf["lat"] * e["lat"]
        assert_representation_allclose(s_lat2, s_lat)

        s_distance = s + 1.0 * u.pc * e["distance"]
        assert_quantity_allclose(s_distance.lon, s.lon, atol=1e-10 * u.rad)
        assert_quantity_allclose(s_distance.lat, s.lat, atol=1e-10 * u.rad)
        assert_quantity_allclose(s_distance.distance, s.distance + 1.0 * u.pc)
        s_distance2 = s + 1.0 * u.pc * sf["distance"] * e["distance"]
        assert_representation_allclose(s_distance2, s_distance)

    def test_unit_spherical(self):
        s = UnitSphericalRepresentation(
            lon=[0.0, 6.0, 21.0] * u.hourangle, lat=[0.0, -30.0, 85.0] * u.deg
        )

        e = s.unit_vectors()
        self.check_unit_vectors(e)
        sf = s.scale_factors()
        self.check_scale_factors(sf, s)

        s_lon = s + 1e-5 * np.cos(s.lat) * e["lon"]
        assert_quantity_allclose(s_lon.lon, s.lon + 1e-5 * u.rad, atol=1e-10 * u.rad)
        assert_quantity_allclose(s_lon.lat, s.lat, atol=1e-10 * u.rad)
        s_lon2 = s + 1e-5 * u.radian * sf["lon"] * e["lon"]
        assert_representation_allclose(s_lon2, s_lon)

        s_lat = s + 1e-5 * e["lat"]
        assert_quantity_allclose(s_lat.lon, s.lon)
        assert_quantity_allclose(s_lat.lat, s.lat + 1e-5 * u.rad, atol=1e-10 * u.rad)
        s_lat2 = s + 1.0e-5 * u.radian * sf["lat"] * e["lat"]
        assert_representation_allclose(s_lat2, s_lat)

    def test_radial(self):
        r = RadialRepresentation(10.0 * u.kpc)
        with pytest.raises(NotImplementedError):
            r.unit_vectors()
        sf = r.scale_factors()
        assert np.all(sf["distance"] == 1.0 * u.one)
        assert np.all(r.norm() == r.distance)
        with pytest.raises(TypeError):
            r + r

    def test_physical_spherical(self):
        s = PhysicsSphericalRepresentation(
            phi=[0.0, 6.0, 21.0] * u.hourangle,
            theta=[90.0, 120.0, 5.0] * u.deg,
            r=[1, 2, 3] * u.kpc,
        )

        e = s.unit_vectors()
        self.check_unit_vectors(e)
        sf = s.scale_factors()
        self.check_scale_factors(sf, s)

        s_phi = s + s.r * 1e-5 * np.sin(s.theta) * e["phi"]
        assert_quantity_allclose(s_phi.phi, s.phi + 1e-5 * u.rad, atol=1e-10 * u.rad)
        assert_quantity_allclose(s_phi.theta, s.theta, atol=1e-10 * u.rad)
        assert_quantity_allclose(s_phi.r, s.r)
        s_phi2 = s + 1e-5 * u.radian * sf["phi"] * e["phi"]
        assert_representation_allclose(s_phi2, s_phi)

        s_theta = s + s.r * 1e-5 * e["theta"]
        assert_quantity_allclose(s_theta.phi, s.phi)
        assert_quantity_allclose(
            s_theta.theta, s.theta + 1e-5 * u.rad, atol=1e-10 * u.rad
        )
        assert_quantity_allclose(s_theta.r, s.r)
        s_theta2 = s + 1.0e-5 * u.radian * sf["theta"] * e["theta"]
        assert_representation_allclose(s_theta2, s_theta)

        s_r = s + 1.0 * u.pc * e["r"]
        assert_quantity_allclose(s_r.phi, s.phi, atol=1e-10 * u.rad)
        assert_quantity_allclose(s_r.theta, s.theta, atol=1e-10 * u.rad)
        assert_quantity_allclose(s_r.r, s.r + 1.0 * u.pc)
        s_r2 = s + 1.0 * u.pc * sf["r"] * e["r"]
        assert_representation_allclose(s_r2, s_r)

    def test_cartesian(self):
        s = CartesianRepresentation(
            x=[1, 2, 3] * u.pc, y=[2, 3, 4] * u.Mpc, z=[3, 4, 5] * u.kpc
        )

        e = s.unit_vectors()
        sf = s.scale_factors()
        for v, expected in zip(
            e.values(),
            ([1.0, 0.0, 0.0] * u.one, [0.0, 1.0, 0.0] * u.one, [0.0, 0.0, 1.0] * u.one),
        ):
            assert np.all(v.get_xyz(xyz_axis=-1) == expected)
        for f in sf.values():
            assert np.all(f == 1.0 * u.one)

    def test_cylindrical(self):
        s = CylindricalRepresentation(
            rho=[1, 2, 3] * u.pc, phi=[0.0, 90.0, -45.0] * u.deg, z=[3, 4, 5] * u.kpc
        )
        e = s.unit_vectors()
        self.check_unit_vectors(e)
        sf = s.scale_factors()
        self.check_scale_factors(sf, s)

        s_rho = s + 1.0 * u.pc * e["rho"]
        assert_quantity_allclose(s_rho.rho, s.rho + 1.0 * u.pc)
        assert_quantity_allclose(s_rho.phi, s.phi)
        assert_quantity_allclose(s_rho.z, s.z)
        s_rho2 = s + 1.0 * u.pc * sf["rho"] * e["rho"]
        assert_representation_allclose(s_rho2, s_rho)

        s_phi = s + s.rho * 1e-5 * e["phi"]
        assert_quantity_allclose(s_phi.rho, s.rho)
        assert_quantity_allclose(s_phi.phi, s.phi + 1e-5 * u.rad)
        assert_quantity_allclose(s_phi.z, s.z)
        s_phi2 = s + 1e-5 * u.radian * sf["phi"] * e["phi"]
        assert_representation_allclose(s_phi2, s_phi)

        s_z = s + 1.0 * u.pc * e["z"]
        assert_quantity_allclose(s_z.rho, s.rho)
        assert_quantity_allclose(s_z.phi, s.phi, atol=1e-10 * u.rad)
        assert_quantity_allclose(s_z.z, s.z + 1.0 * u.pc)
        s_z2 = s + 1.0 * u.pc * sf["z"] * e["z"]
        assert_representation_allclose(s_z2, s_z)


@pytest.mark.parametrize("omit_coslat", [False, True], scope="class")
class TestSphericalDifferential:
    # these test cases are subclassed for SphericalCosLatDifferential,
    # hence some tests depend on omit_coslat.

    def _setup(self, omit_coslat):
        if omit_coslat:
            self.SD_cls = SphericalCosLatDifferential
        else:
            self.SD_cls = SphericalDifferential

        s = SphericalRepresentation(
            lon=[0.0, 6.0, 21.0] * u.hourangle,
            lat=[0.0, -30.0, 85.0] * u.deg,
            distance=[1, 2, 3] * u.kpc,
        )
        self.s = s
        self.e = s.unit_vectors()
        self.sf = s.scale_factors(omit_coslat=omit_coslat)

    def test_name_coslat(self, omit_coslat):
        self._setup(omit_coslat)
        if omit_coslat:
            assert self.SD_cls is SphericalCosLatDifferential
            assert self.SD_cls.name == "sphericalcoslat"
        else:
            assert self.SD_cls is SphericalDifferential
            assert self.SD_cls.name == "spherical"
        assert self.SD_cls.name in DIFFERENTIAL_CLASSES

    def test_simple_differentials(self, omit_coslat):
        self._setup(omit_coslat)
        s, e, sf = self.s, self.e, self.sf

        o_lon = self.SD_cls(1.0 * u.arcsec, 0.0 * u.arcsec, 0.0 * u.kpc)
        o_lonc = o_lon.to_cartesian(base=s)
        o_lon2 = self.SD_cls.from_cartesian(o_lonc, base=s)
        assert_differential_allclose(o_lon, o_lon2)
        # simple check by hand for first element.
        # lat[0] is 0, so cos(lat) term doesn't matter.
        assert_quantity_allclose(
            o_lonc[0].xyz, [0.0, np.pi / 180.0 / 3600.0, 0.0] * u.kpc
        )
        # check all using unit vectors and scale factors.
        s_lon = s + 1.0 * u.arcsec * sf["lon"] * e["lon"]
        assert_representation_allclose(o_lonc, s_lon - s, atol=1 * u.npc)
        s_lon2 = s + o_lon
        assert_representation_allclose(s_lon2, s_lon, atol=1 * u.npc)

        o_lat = self.SD_cls(0.0 * u.arcsec, 1.0 * u.arcsec, 0.0 * u.kpc)
        o_latc = o_lat.to_cartesian(base=s)
        assert_quantity_allclose(
            o_latc[0].xyz, [0.0, 0.0, np.pi / 180.0 / 3600.0] * u.kpc, atol=1.0 * u.npc
        )
        s_lat = s + 1.0 * u.arcsec * sf["lat"] * e["lat"]
        assert_representation_allclose(o_latc, s_lat - s, atol=1 * u.npc)
        s_lat2 = s + o_lat
        assert_representation_allclose(s_lat2, s_lat, atol=1 * u.npc)

        o_distance = self.SD_cls(0.0 * u.arcsec, 0.0 * u.arcsec, 1.0 * u.mpc)
        o_distancec = o_distance.to_cartesian(base=s)
        assert_quantity_allclose(
            o_distancec[0].xyz, [1e-6, 0.0, 0.0] * u.kpc, atol=1.0 * u.npc
        )
        s_distance = s + 1.0 * u.mpc * sf["distance"] * e["distance"]
        assert_representation_allclose(o_distancec, s_distance - s, atol=1 * u.npc)
        s_distance2 = s + o_distance
        assert_representation_allclose(s_distance2, s_distance)

    def test_differential_arithmetic(self, omit_coslat):
        self._setup(omit_coslat)
        s = self.s

        o_lon = self.SD_cls(1.0 * u.arcsec, 0.0 * u.arcsec, 0.0 * u.kpc)
        o_lon_by_2 = o_lon / 2.0
        assert_representation_allclose(
            o_lon_by_2.to_cartesian(s) * 2.0, o_lon.to_cartesian(s), atol=1e-10 * u.kpc
        )
        assert_representation_allclose(
            s + o_lon, s + 2 * o_lon_by_2, atol=1e-10 * u.kpc
        )
        o_lon_rec = o_lon_by_2 + o_lon_by_2
        assert_representation_allclose(s + o_lon, s + o_lon_rec, atol=1e-10 * u.kpc)
        o_lon_0 = o_lon - o_lon
        for c in o_lon_0.components:
            assert np.all(getattr(o_lon_0, c) == 0.0)
        o_lon2 = self.SD_cls(1 * u.mas / u.yr, 0 * u.mas / u.yr, 0 * u.km / u.s)
        assert_quantity_allclose(
            o_lon2.norm(s)[0], 4.74 * u.km / u.s, atol=0.01 * u.km / u.s
        )
        assert_representation_allclose(
            o_lon2.to_cartesian(s) * 1000.0 * u.yr,
            o_lon.to_cartesian(s),
            atol=1e-10 * u.kpc,
        )
        s_off = s + o_lon
        s_off2 = s + o_lon2 * 1000.0 * u.yr
        assert_representation_allclose(s_off, s_off2, atol=1e-10 * u.kpc)

        factor = 1e5 * u.radian / u.arcsec
        if not omit_coslat:
            factor = factor / np.cos(s.lat)
        s_off_big = s + o_lon * factor

        assert_representation_allclose(
            s_off_big,
            SphericalRepresentation(
                s.lon + 90.0 * u.deg, 0.0 * u.deg, 1e5 * s.distance
            ),
            atol=5.0 * u.kpc,
        )

        o_lon3c = CartesianRepresentation(0.0, 4.74047, 0.0, unit=u.km / u.s)
        o_lon3 = self.SD_cls.from_cartesian(o_lon3c, base=s)
        expected0 = self.SD_cls(
            1.0 * u.mas / u.yr, 0.0 * u.mas / u.yr, 0.0 * u.km / u.s
        )
        assert_differential_allclose(o_lon3[0], expected0)
        s_off_big2 = s + o_lon3 * 1e5 * u.yr * u.radian / u.mas
        assert_representation_allclose(
            s_off_big2,
            SphericalRepresentation(90.0 * u.deg, 0.0 * u.deg, 1e5 * u.kpc),
            atol=5.0 * u.kpc,
        )

        with pytest.raises(TypeError):
            o_lon - s
        with pytest.raises(TypeError):
            s.to_cartesian() + o_lon

    def test_differential_init_errors(self, omit_coslat):
        self._setup(omit_coslat)
        s = self.s
        with pytest.raises(u.UnitsError):
            self.SD_cls(1.0 * u.arcsec, 0.0, 0.0)
        with pytest.raises(TypeError):
            self.SD_cls(1.0 * u.arcsec, 0.0 * u.arcsec, 0.0 * u.kpc, False, False)
        with pytest.raises(TypeError):
            self.SD_cls(
                1.0 * u.arcsec,
                0.0 * u.arcsec,
                0.0 * u.kpc,
                copy=False,
                d_lat=0.0 * u.arcsec,
            )
        with pytest.raises(TypeError):
            self.SD_cls(
                1.0 * u.arcsec, 0.0 * u.arcsec, 0.0 * u.kpc, copy=False, flying="circus"
            )
        with pytest.raises(ValueError):
            self.SD_cls(
                np.ones(2) * u.arcsec, np.zeros(3) * u.arcsec, np.zeros(2) * u.kpc
            )
        with pytest.raises(u.UnitsError):
            self.SD_cls(1.0 * u.arcsec, 1.0 * u.s, 0.0 * u.kpc)
        with pytest.raises(u.UnitsError):
            self.SD_cls(1.0 * u.kpc, 1.0 * u.arcsec, 0.0 * u.kpc)
        o = self.SD_cls(1.0 * u.arcsec, 1.0 * u.arcsec, 0.0 * u.km / u.s)
        with pytest.raises(u.UnitsError):
            o.to_cartesian(s)
        with pytest.raises(AttributeError):
            o.d_lat = 0.0 * u.arcsec
        with pytest.raises(AttributeError):
            del o.d_lat

        o = self.SD_cls(1.0 * u.arcsec, 1.0 * u.arcsec, 0.0 * u.km)
        with pytest.raises(TypeError):
            o.to_cartesian()
        c = CartesianRepresentation(10.0, 0.0, 0.0, unit=u.km)
        with pytest.raises(TypeError):
            self.SD_cls.to_cartesian(c)
        with pytest.raises(TypeError):
            self.SD_cls.from_cartesian(c)
        with pytest.raises(TypeError):
            self.SD_cls.from_cartesian(c, SphericalRepresentation)


@pytest.mark.parametrize("omit_coslat", [False, True], scope="class")
class TestUnitSphericalDifferential:
    def _setup(self, omit_coslat):
        if omit_coslat:
            self.USD_cls = UnitSphericalCosLatDifferential
        else:
            self.USD_cls = UnitSphericalDifferential

        s = UnitSphericalRepresentation(
            lon=[0.0, 6.0, 21.0] * u.hourangle, lat=[0.0, -30.0, 85.0] * u.deg
        )
        self.s = s
        self.e = s.unit_vectors()
        self.sf = s.scale_factors(omit_coslat=omit_coslat)

    def test_name_coslat(self, omit_coslat):
        self._setup(omit_coslat)
        if omit_coslat:
            assert self.USD_cls is UnitSphericalCosLatDifferential
            assert self.USD_cls.name == "unitsphericalcoslat"
        else:
            assert self.USD_cls is UnitSphericalDifferential
            assert self.USD_cls.name == "unitspherical"
        assert self.USD_cls.name in DIFFERENTIAL_CLASSES

    def test_simple_differentials(self, omit_coslat):
        self._setup(omit_coslat)
        s, e, sf = self.s, self.e, self.sf

        o_lon = self.USD_cls(1.0 * u.arcsec, 0.0 * u.arcsec)
        o_lonc = o_lon.to_cartesian(base=s)
        o_lon2 = self.USD_cls.from_cartesian(o_lonc, base=s)
        assert_differential_allclose(o_lon, o_lon2)
        # simple check by hand for first element
        # (lat[0]=0, so works for both normal and CosLat differential)
        assert_quantity_allclose(
            o_lonc[0].xyz, [0.0, np.pi / 180.0 / 3600.0, 0.0] * u.one
        )
        # check all using unit vectors and scale factors.
        s_lon = s + 1.0 * u.arcsec * sf["lon"] * e["lon"]
        assert type(s_lon) is SphericalRepresentation
        assert_representation_allclose(o_lonc, s_lon - s, atol=1e-10 * u.one)
        s_lon2 = s + o_lon
        assert_representation_allclose(s_lon2, s_lon, atol=1e-10 * u.one)

        o_lat = self.USD_cls(0.0 * u.arcsec, 1.0 * u.arcsec)
        o_latc = o_lat.to_cartesian(base=s)
        assert_quantity_allclose(
            o_latc[0].xyz,
            [0.0, 0.0, np.pi / 180.0 / 3600.0] * u.one,
            atol=1e-10 * u.one,
        )
        s_lat = s + 1.0 * u.arcsec * sf["lat"] * e["lat"]
        assert type(s_lat) is SphericalRepresentation
        assert_representation_allclose(o_latc, s_lat - s, atol=1e-10 * u.one)
        s_lat2 = s + o_lat
        assert_representation_allclose(s_lat2, s_lat, atol=1e-10 * u.one)

    def test_differential_arithmetic(self, omit_coslat):
        self._setup(omit_coslat)
        s = self.s

        o_lon = self.USD_cls(1.0 * u.arcsec, 0.0 * u.arcsec)
        o_lon_by_2 = o_lon / 2.0
        assert type(o_lon_by_2) is self.USD_cls
        assert_representation_allclose(
            o_lon_by_2.to_cartesian(s) * 2.0, o_lon.to_cartesian(s), atol=1e-10 * u.one
        )
        s_lon = s + o_lon
        s_lon2 = s + 2 * o_lon_by_2
        assert type(s_lon) is SphericalRepresentation
        assert_representation_allclose(s_lon, s_lon2, atol=1e-10 * u.one)
        o_lon_rec = o_lon_by_2 + o_lon_by_2
        assert type(o_lon_rec) is self.USD_cls
        assert representation_equal(o_lon, o_lon_rec)
        assert_representation_allclose(s + o_lon, s + o_lon_rec, atol=1e-10 * u.one)
        o_lon_0 = o_lon - o_lon
        assert type(o_lon_0) is self.USD_cls
        for c in o_lon_0.components:
            assert np.all(getattr(o_lon_0, c) == 0.0)

        o_lon2 = self.USD_cls(1.0 * u.mas / u.yr, 0.0 * u.mas / u.yr)
        kks = u.km / u.kpc / u.s
        assert_quantity_allclose(o_lon2.norm(s)[0], 4.74047 * kks, atol=1e-4 * kks)
        assert_representation_allclose(
            o_lon2.to_cartesian(s) * 1000.0 * u.yr,
            o_lon.to_cartesian(s),
            atol=1e-10 * u.one,
        )
        s_off = s + o_lon
        s_off2 = s + o_lon2 * 1000.0 * u.yr
        assert_representation_allclose(s_off, s_off2, atol=1e-10 * u.one)

        factor = 1e5 * u.radian / u.arcsec
        if not omit_coslat:
            factor = factor / np.cos(s.lat)
        s_off_big = s + o_lon * factor

        assert_representation_allclose(
            s_off_big,
            SphericalRepresentation(s.lon + 90.0 * u.deg, 0.0 * u.deg, 1e5),
            atol=5.0 * u.one,
        )

        o_lon3c = CartesianRepresentation(0.0, 4.74047, 0.0, unit=kks)
        # This looses information!!
        o_lon3 = self.USD_cls.from_cartesian(o_lon3c, base=s)
        expected0 = self.USD_cls(1.0 * u.mas / u.yr, 0.0 * u.mas / u.yr)
        assert_differential_allclose(o_lon3[0], expected0)
        # Part of motion kept.
        part_kept = s.cross(CartesianRepresentation(0, 1, 0, unit=u.one)).norm()
        assert_quantity_allclose(
            o_lon3.norm(s), 4.74047 * part_kept * kks, atol=1e-10 * kks
        )
        # (lat[0]=0, so works for both normal and CosLat differential)
        s_off_big2 = s + o_lon3 * 1e5 * u.yr * u.radian / u.mas
        expected0 = SphericalRepresentation(90.0 * u.deg, 0.0 * u.deg, 1e5 * u.one)
        assert_representation_allclose(s_off_big2[0], expected0, atol=5.0 * u.one)

    def test_differential_init_errors(self, omit_coslat):
        self._setup(omit_coslat)
        with pytest.raises(u.UnitsError):
            self.USD_cls(0.0 * u.deg, 10.0 * u.deg / u.yr)


class TestRadialDifferential:
    def setup_method(self):
        s = SphericalRepresentation(
            lon=[0.0, 6.0, 21.0] * u.hourangle,
            lat=[0.0, -30.0, 85.0] * u.deg,
            distance=[1, 2, 3] * u.kpc,
        )
        self.s = s
        self.r = s.represent_as(RadialRepresentation)
        self.e = s.unit_vectors()
        self.sf = s.scale_factors()

    def test_name(self):
        assert RadialDifferential.name == "radial"
        assert RadialDifferential.name in DIFFERENTIAL_CLASSES

    def test_simple_differentials(self):
        r, s, e, sf = self.r, self.s, self.e, self.sf

        o_distance = RadialDifferential(1.0 * u.mpc)
        # Can be applied to RadialRepresentation, though not most useful.
        r_distance = r + o_distance
        assert_quantity_allclose(
            r_distance.distance, r.distance + o_distance.d_distance
        )
        r_distance2 = o_distance + r
        assert_quantity_allclose(
            r_distance2.distance, r.distance + o_distance.d_distance
        )
        # More sense to apply it relative to spherical representation.
        o_distancec = o_distance.to_cartesian(base=s)
        assert_quantity_allclose(
            o_distancec[0].xyz, [1e-6, 0.0, 0.0] * u.kpc, atol=1.0 * u.npc
        )
        o_recover = RadialDifferential.from_cartesian(o_distancec, base=s)
        assert_quantity_allclose(o_recover.d_distance, o_distance.d_distance)

        s_distance = s + 1.0 * u.mpc * sf["distance"] * e["distance"]
        assert_representation_allclose(o_distancec, s_distance - s, atol=1 * u.npc)
        s_distance2 = s + o_distance
        assert_representation_allclose(s_distance2, s_distance)


class TestPhysicsSphericalDifferential:
    """Test copied from SphericalDifferential, so less extensive."""

    def setup_method(self):
        s = PhysicsSphericalRepresentation(
            phi=[0.0, 90.0, 315.0] * u.deg,
            theta=[90.0, 120.0, 5.0] * u.deg,
            r=[1, 2, 3] * u.kpc,
        )
        self.s = s
        self.e = s.unit_vectors()
        self.sf = s.scale_factors()

    def test_name(self):
        assert PhysicsSphericalDifferential.name == "physicsspherical"
        assert PhysicsSphericalDifferential.name in DIFFERENTIAL_CLASSES

    def test_simple_differentials(self):
        s, e, sf = self.s, self.e, self.sf

        o_phi = PhysicsSphericalDifferential(1 * u.arcsec, 0 * u.arcsec, 0 * u.kpc)
        o_phic = o_phi.to_cartesian(base=s)
        o_phi2 = PhysicsSphericalDifferential.from_cartesian(o_phic, base=s)
        assert_quantity_allclose(o_phi.d_phi, o_phi2.d_phi, atol=1.0 * u.narcsec)
        assert_quantity_allclose(o_phi.d_theta, o_phi2.d_theta, atol=1.0 * u.narcsec)
        assert_quantity_allclose(o_phi.d_r, o_phi2.d_r, atol=1.0 * u.npc)
        # simple check by hand for first element.
        assert_quantity_allclose(
            o_phic[0].xyz, [0.0, np.pi / 180.0 / 3600.0, 0.0] * u.kpc, atol=1.0 * u.npc
        )
        # check all using unit vectors and scale factors.
        s_phi = s + 1.0 * u.arcsec * sf["phi"] * e["phi"]
        assert_representation_allclose(o_phic, s_phi - s, atol=1e-10 * u.kpc)

        o_theta = PhysicsSphericalDifferential(0 * u.arcsec, 1 * u.arcsec, 0 * u.kpc)
        o_thetac = o_theta.to_cartesian(base=s)
        assert_quantity_allclose(
            o_thetac[0].xyz,
            [0.0, 0.0, -np.pi / 180.0 / 3600.0] * u.kpc,
            atol=1.0 * u.npc,
        )
        s_theta = s + 1.0 * u.arcsec * sf["theta"] * e["theta"]
        assert_representation_allclose(o_thetac, s_theta - s, atol=1e-10 * u.kpc)
        s_theta2 = s + o_theta
        assert_representation_allclose(s_theta2, s_theta, atol=1e-10 * u.kpc)

        o_r = PhysicsSphericalDifferential(0 * u.arcsec, 0 * u.arcsec, 1 * u.mpc)
        o_rc = o_r.to_cartesian(base=s)
        assert_quantity_allclose(
            o_rc[0].xyz, [1e-6, 0.0, 0.0] * u.kpc, atol=1.0 * u.npc
        )
        s_r = s + 1.0 * u.mpc * sf["r"] * e["r"]
        assert_representation_allclose(o_rc, s_r - s, atol=1e-10 * u.kpc)
        s_r2 = s + o_r
        assert_representation_allclose(s_r2, s_r)

    def test_differential_init_errors(self):
        with pytest.raises(u.UnitsError):
            PhysicsSphericalDifferential(1.0 * u.arcsec, 0.0, 0.0)


class TestCylindricalDifferential:
    """Test copied from SphericalDifferential, so less extensive."""

    def setup_method(self):
        s = CylindricalRepresentation(
            rho=[1, 2, 3] * u.kpc, phi=[0.0, 90.0, 315.0] * u.deg, z=[3, 2, 1] * u.kpc
        )
        self.s = s
        self.e = s.unit_vectors()
        self.sf = s.scale_factors()

    def test_name(self):
        assert CylindricalDifferential.name == "cylindrical"
        assert CylindricalDifferential.name in DIFFERENTIAL_CLASSES

    def test_simple_differentials(self):
        s, e, sf = self.s, self.e, self.sf

        o_rho = CylindricalDifferential(1.0 * u.mpc, 0.0 * u.arcsec, 0.0 * u.kpc)
        o_rhoc = o_rho.to_cartesian(base=s)
        assert_quantity_allclose(o_rhoc[0].xyz, [1.0e-6, 0.0, 0.0] * u.kpc)
        s_rho = s + 1.0 * u.mpc * sf["rho"] * e["rho"]
        assert_representation_allclose(o_rhoc, s_rho - s, atol=1e-10 * u.kpc)
        s_rho2 = s + o_rho
        assert_representation_allclose(s_rho2, s_rho)

        o_phi = CylindricalDifferential(0.0 * u.kpc, 1.0 * u.arcsec, 0.0 * u.kpc)
        o_phic = o_phi.to_cartesian(base=s)
        o_phi2 = CylindricalDifferential.from_cartesian(o_phic, base=s)
        assert_quantity_allclose(o_phi.d_rho, o_phi2.d_rho, atol=1.0 * u.npc)
        assert_quantity_allclose(o_phi.d_phi, o_phi2.d_phi, atol=1.0 * u.narcsec)
        assert_quantity_allclose(o_phi.d_z, o_phi2.d_z, atol=1.0 * u.npc)
        # simple check by hand for first element.
        assert_quantity_allclose(
            o_phic[0].xyz, [0.0, np.pi / 180.0 / 3600.0, 0.0] * u.kpc
        )
        # check all using unit vectors and scale factors.
        s_phi = s + 1.0 * u.arcsec * sf["phi"] * e["phi"]
        assert_representation_allclose(o_phic, s_phi - s, atol=1e-10 * u.kpc)

        o_z = CylindricalDifferential(0.0 * u.kpc, 0.0 * u.arcsec, 1.0 * u.mpc)
        o_zc = o_z.to_cartesian(base=s)
        assert_quantity_allclose(o_zc[0].xyz, [0.0, 0.0, 1.0e-6] * u.kpc)
        s_z = s + 1.0 * u.mpc * sf["z"] * e["z"]
        assert_representation_allclose(o_zc, s_z - s, atol=1e-10 * u.kpc)
        s_z2 = s + o_z
        assert_representation_allclose(s_z2, s_z)

    def test_differential_init_errors(self):
        with pytest.raises(u.UnitsError):
            CylindricalDifferential(1.0 * u.pc, 1.0 * u.arcsec, 3.0 * u.km / u.s)


class TestCartesianDifferential:
    """Test copied from SphericalDifferential, so less extensive."""

    def setup_method(self):
        s = CartesianRepresentation(
            x=[1, 2, 3] * u.kpc, y=[2, 3, 1] * u.kpc, z=[3, 1, 2] * u.kpc
        )
        self.s = s
        self.e = s.unit_vectors()
        self.sf = s.scale_factors()

    def test_name(self):
        assert CartesianDifferential.name == "cartesian"
        assert CartesianDifferential.name in DIFFERENTIAL_CLASSES

    def test_simple_differentials(self):
        s, e, sf = self.s, self.e, self.sf

        for d, differential in (  # test different inits while we're at it.
            ("x", CartesianDifferential(1.0 * u.pc, 0.0 * u.pc, 0.0 * u.pc)),
            ("y", CartesianDifferential([0.0, 1.0, 0.0], unit=u.pc)),
            (
                "z",
                CartesianDifferential(np.array([[0.0, 0.0, 1.0]]) * u.pc, xyz_axis=1),
            ),
        ):
            o_c = differential.to_cartesian(base=s)
            o_c2 = differential.to_cartesian()
            assert np.all(representation_equal(o_c, o_c2))
            assert all(
                np.all(getattr(differential, "d_" + c) == getattr(o_c, c))
                for c in ("x", "y", "z")
            )
            differential2 = CartesianDifferential.from_cartesian(o_c)
            assert np.all(representation_equal(differential2, differential))
            differential3 = CartesianDifferential.from_cartesian(o_c, base=o_c)
            assert np.all(representation_equal(differential3, differential))

            s_off = s + 1.0 * u.pc * sf[d] * e[d]
            assert_representation_allclose(o_c, s_off - s, atol=1e-10 * u.kpc)
            s_off2 = s + differential
            assert_representation_allclose(s_off2, s_off)

    def test_init_failures(self):
        with pytest.raises(ValueError):
            CartesianDifferential(1.0 * u.kpc / u.s, 2.0 * u.kpc)
        with pytest.raises(u.UnitsError):
            CartesianDifferential(1.0 * u.kpc / u.s, 2.0 * u.kpc, 3.0 * u.kpc)
        with pytest.raises(ValueError):
            CartesianDifferential(1.0 * u.kpc, 2.0 * u.kpc, 3.0 * u.kpc, xyz_axis=1)


class TestDifferentialConversion:
    def setup_method(self):
        self.s = SphericalRepresentation(
            lon=[0.0, 6.0, 21.0] * u.hourangle,
            lat=[0.0, -30.0, 85.0] * u.deg,
            distance=[1, 2, 3] * u.kpc,
        )

    @pytest.mark.parametrize(
        "sd_cls", [SphericalDifferential, SphericalCosLatDifferential]
    )
    def test_represent_as_own_class(self, sd_cls):
        so = sd_cls(1.0 * u.deg, 2.0 * u.deg, 0.1 * u.kpc)
        so2 = so.represent_as(sd_cls)
        assert so2 is so

    def test_represent_other_coslat(self):
        s = self.s
        coslat = np.cos(s.lat)
        so = SphericalDifferential(1.0 * u.deg, 2.0 * u.deg, 0.1 * u.kpc)
        so_coslat = so.represent_as(SphericalCosLatDifferential, base=s)
        assert_quantity_allclose(so.d_lon * coslat, so_coslat.d_lon_coslat)
        so2 = so_coslat.represent_as(SphericalDifferential, base=s)
        assert np.all(representation_equal(so2, so))
        so3 = SphericalDifferential.from_representation(so_coslat, base=s)
        assert np.all(representation_equal(so3, so))
        so_coslat2 = SphericalCosLatDifferential.from_representation(so, base=s)
        assert np.all(representation_equal(so_coslat2, so_coslat))
        # Also test UnitSpherical
        us = s.represent_as(UnitSphericalRepresentation)
        uo = so.represent_as(UnitSphericalDifferential)
        uo_coslat = so.represent_as(UnitSphericalCosLatDifferential, base=s)
        assert_quantity_allclose(uo.d_lon * coslat, uo_coslat.d_lon_coslat)
        uo2 = uo_coslat.represent_as(UnitSphericalDifferential, base=us)
        assert np.all(representation_equal(uo2, uo))
        uo3 = UnitSphericalDifferential.from_representation(uo_coslat, base=us)
        assert np.all(representation_equal(uo3, uo))
        uo_coslat2 = UnitSphericalCosLatDifferential.from_representation(uo, base=us)
        assert np.all(representation_equal(uo_coslat2, uo_coslat))
        uo_coslat3 = uo.represent_as(UnitSphericalCosLatDifferential, base=us)
        assert np.all(representation_equal(uo_coslat3, uo_coslat))

    @pytest.mark.parametrize(
        "sd_cls", [SphericalDifferential, SphericalCosLatDifferential]
    )
    @pytest.mark.parametrize(
        "r_cls",
        (
            SphericalRepresentation,
            UnitSphericalRepresentation,
            PhysicsSphericalRepresentation,
            CylindricalRepresentation,
        ),
    )
    def test_represent_regular_class(self, sd_cls, r_cls):
        so = sd_cls(1.0 * u.deg, 2.0 * u.deg, 0.1 * u.kpc)
        r = so.represent_as(r_cls, base=self.s)
        c = so.to_cartesian(self.s)
        r_check = c.represent_as(r_cls)
        assert np.all(representation_equal(r, r_check))
        so2 = sd_cls.from_representation(r, base=self.s)
        so3 = sd_cls.from_cartesian(r.to_cartesian(), self.s)
        assert np.all(representation_equal(so2, so3))

    @pytest.mark.parametrize(
        "sd_cls", [SphericalDifferential, SphericalCosLatDifferential]
    )
    def test_convert_physics(self, sd_cls):
        # Conversion needs no base for SphericalDifferential, but does
        # need one (to get the latitude) for SphericalCosLatDifferential.
        if sd_cls is SphericalDifferential:
            usd_cls = UnitSphericalDifferential
            base_s = base_u = base_p = None
        else:
            usd_cls = UnitSphericalCosLatDifferential
            base_s = self.s[1]
            base_u = base_s.represent_as(UnitSphericalRepresentation)
            base_p = base_s.represent_as(PhysicsSphericalRepresentation)

        so = sd_cls(1.0 * u.deg, 2.0 * u.deg, 0.1 * u.kpc)
        po = so.represent_as(PhysicsSphericalDifferential, base=base_s)
        so2 = sd_cls.from_representation(po, base=base_s)
        assert_differential_allclose(so, so2)
        po2 = PhysicsSphericalDifferential.from_representation(so, base=base_p)
        assert_differential_allclose(po, po2)
        so3 = po.represent_as(sd_cls, base=base_p)
        assert_differential_allclose(so, so3)

        s = self.s
        p = s.represent_as(PhysicsSphericalRepresentation)
        cso = so.to_cartesian(s[1])
        cpo = po.to_cartesian(p[1])
        assert_representation_allclose(cso, cpo)
        assert_representation_allclose(s[1] + so, p[1] + po)
        po2 = so.represent_as(
            PhysicsSphericalDifferential, base=None if base_s is None else s
        )
        assert_representation_allclose(s + so, p + po2)

        suo = usd_cls.from_representation(so)
        puo = usd_cls.from_representation(po, base=base_u)
        assert_differential_allclose(suo, puo)
        suo2 = so.represent_as(usd_cls)
        puo2 = po.represent_as(usd_cls, base=base_p)
        assert_differential_allclose(suo2, puo2)
        assert_differential_allclose(puo, puo2)

        sro = RadialDifferential.from_representation(so)
        pro = RadialDifferential.from_representation(po)
        assert representation_equal(sro, pro)
        sro2 = so.represent_as(RadialDifferential)
        pro2 = po.represent_as(RadialDifferential)
        assert representation_equal(sro2, pro2)
        assert representation_equal(pro, pro2)

    @pytest.mark.parametrize(
        ("sd_cls", "usd_cls"),
        [
            (SphericalDifferential, UnitSphericalDifferential),
            (SphericalCosLatDifferential, UnitSphericalCosLatDifferential),
        ],
    )
    def test_convert_unit_spherical_radial(self, sd_cls, usd_cls):
        s = self.s
        us = s.represent_as(UnitSphericalRepresentation)
        rs = s.represent_as(RadialRepresentation)
        assert_representation_allclose(rs * us, s)

        uo = usd_cls(2.0 * u.deg, 1.0 * u.deg)
        so = uo.represent_as(sd_cls, base=s)
        assert_quantity_allclose(so.d_distance, 0.0 * u.kpc, atol=1.0 * u.npc)
        uo2 = so.represent_as(usd_cls)
        assert_representation_allclose(uo.to_cartesian(us), uo2.to_cartesian(us))
        so1 = sd_cls(2.0 * u.deg, 1.0 * u.deg, 5.0 * u.pc)
        uo_r = so1.represent_as(usd_cls)
        ro_r = so1.represent_as(RadialDifferential)
        assert np.all(representation_equal(uo_r, uo))
        assert np.all(representation_equal(ro_r, RadialDifferential(5.0 * u.pc)))

    @pytest.mark.parametrize(
        "sd_cls", [SphericalDifferential, SphericalCosLatDifferential]
    )
    def test_convert_cylindrial(self, sd_cls):
        s = self.s
        so = sd_cls(1.0 * u.deg, 2.0 * u.deg, 0.1 * u.kpc)
        cyo = so.represent_as(CylindricalDifferential, base=s)
        cy = s.represent_as(CylindricalRepresentation)
        so1 = cyo.represent_as(sd_cls, base=cy)
        assert_representation_allclose(so.to_cartesian(s), so1.to_cartesian(s))
        cyo2 = CylindricalDifferential.from_representation(so, base=cy)
        assert_representation_allclose(
            cyo2.to_cartesian(base=cy), cyo.to_cartesian(base=cy)
        )
        so2 = sd_cls.from_representation(cyo2, base=s)
        assert_representation_allclose(so.to_cartesian(s), so2.to_cartesian(s))

    @pytest.mark.parametrize(
        "sd_cls", [SphericalDifferential, SphericalCosLatDifferential]
    )
    def test_combinations(self, sd_cls):
        if sd_cls is SphericalDifferential:
            uo = UnitSphericalDifferential(2.0 * u.deg, 1.0 * u.deg)
            uo_d_lon = uo.d_lon
        else:
            uo = UnitSphericalCosLatDifferential(2.0 * u.deg, 1.0 * u.deg)
            uo_d_lon = uo.d_lon_coslat
        ro = RadialDifferential(1.0 * u.mpc)
        so1 = uo + ro
        so1c = sd_cls(uo_d_lon, uo.d_lat, ro.d_distance)
        assert np.all(representation_equal(so1, so1c))

        so2 = uo - ro
        so2c = sd_cls(uo_d_lon, uo.d_lat, -ro.d_distance)
        assert np.all(representation_equal(so2, so2c))
        so3 = so2 + ro
        so3c = sd_cls(uo_d_lon, uo.d_lat, 0.0 * u.kpc)
        assert np.all(representation_equal(so3, so3c))
        so4 = so1 + ro
        so4c = sd_cls(uo_d_lon, uo.d_lat, 2 * ro.d_distance)
        assert np.all(representation_equal(so4, so4c))
        so5 = so1 - uo
        so5c = sd_cls(0 * u.deg, 0.0 * u.deg, ro.d_distance)
        assert np.all(representation_equal(so5, so5c))
        assert_representation_allclose(self.s + (uo + ro), self.s + so1)


@pytest.mark.parametrize(
    "op,args",
    [
        (operator.neg, ()),
        (operator.pos, ()),
        (operator.mul, (-8.0,)),
        (operator.truediv, ([4.0, 8.0] * u.s,)),
    ],
    scope="class",
)
class TestArithmeticWithDifferentials:
    def setup_class(self):
        self.cr = CartesianRepresentation([1, 2, 3] * u.kpc)
        self.cd = CartesianDifferential([0.1, -0.2, 0.3] * u.km / u.s)
        self.c = self.cr.with_differentials(self.cd)

    def test_operation_cartesian(self, op, args):
        ncr = op(self.c, *args)
        expected = (op(self.cr, *args)).with_differentials(op(self.cd, *args))
        assert np.all(ncr == expected)

    def test_operation_radial(self, op, args):
        rep = self.c.represent_as(RadialRepresentation, {"s": RadialDifferential})
        result = op(rep, *args)
        expected_distance = op(self.cr.norm(), *args)
        expected_rv = op((self.cr / self.cr.norm()).dot(self.cd), *args)
        assert u.allclose(result.distance, expected_distance)
        assert u.allclose(result.differentials["s"].d_distance, expected_rv)

    @pytest.mark.parametrize(
        "diff_cls",
        [
            SphericalDifferential,
            SphericalCosLatDifferential,
            PhysicsSphericalDifferential,
            CylindricalDifferential,
        ],
    )
    def test_operation_other(self, diff_cls, op, args):
        rep_cls = diff_cls.base_representation
        rep = self.c.represent_as(rep_cls, {"s": diff_cls})
        result = op(rep, *args)

        expected_c = op(self.c, *args)
        expected = expected_c.represent_as(rep_cls, {"s": diff_cls})
        # Check that we match in the representation itself.
        assert_representation_allclose(result, expected)
        assert_differential_allclose(
            result.differentials["s"], expected.differentials["s"]
        )

        # Check that we compare correctly in cartesian as well, just to be sure.
        result_c = result.represent_as(
            CartesianRepresentation, {"s": CartesianDifferential}
        )
        assert_representation_allclose(result_c, expected_c)
        assert_differential_allclose(
            result_c.differentials["s"], expected_c.differentials["s"]
        )

    @pytest.mark.parametrize(
        "rep_cls",
        [
            SphericalRepresentation,
            PhysicsSphericalRepresentation,
            CylindricalRepresentation,
        ],
    )
    def test_operation_cartesian_differential(self, rep_cls, op, args):
        rep = self.c.represent_as(rep_cls, {"s": CartesianDifferential})
        result = op(rep, *args)

        expected_c = op(self.c, *args)
        expected = expected_c.represent_as(rep_cls, {"s": CartesianDifferential})
        # Check that we match in the representation itself.
        assert_representation_allclose(result, expected)
        assert_differential_allclose(
            result.differentials["s"], expected.differentials["s"]
        )

    @pytest.mark.parametrize(
        "diff_cls", [UnitSphericalDifferential, UnitSphericalCosLatDifferential]
    )
    def test_operation_unit_spherical(self, diff_cls, op, args):
        rep_cls = diff_cls.base_representation
        rep = self.c.represent_as(rep_cls, {"s": diff_cls})
        result = op(rep, *args)

        if op not in (operator.neg, operator.pos):
            expected_cls = rep._dimensional_representation
        else:
            expected_cls = rep_cls

        assert type(result) is expected_cls
        assert type(result.differentials["s"]) is diff_cls

        # Have lost information, so unlike above we convert our initial
        # unit-spherical back to Cartesian, and check that applying
        # the operation on that cartesian representation gives the same result.
        # We do not compare the output directly, since for multiplication
        # and division there will be sign flips in the spherical distance.
        expected_c = op(
            rep.represent_as(CartesianRepresentation, {"s": CartesianDifferential}),
            *args,
        )
        result_c = result.represent_as(
            CartesianRepresentation, {"s": CartesianDifferential}
        )
        assert_representation_allclose(result_c, expected_c)
        assert_differential_allclose(
            result_c.differentials["s"], expected_c.differentials["s"]
        )

    @pytest.mark.parametrize(
        "diff_cls",
        [
            RadialDifferential,
            UnitSphericalDifferential,
            UnitSphericalCosLatDifferential,
        ],
    )
    def test_operation_spherical_with_rv_or_pm(self, diff_cls, op, args):
        rep = self.c.represent_as(SphericalRepresentation, {"s": diff_cls})
        result = op(rep, *args)
        assert type(result) is SphericalRepresentation
        assert type(result.differentials["s"]) is diff_cls

        expected_c = op(
            rep.represent_as(CartesianRepresentation, {"s": CartesianDifferential}),
            *args,
        )
        result_c = result.represent_as(
            CartesianRepresentation, {"s": CartesianDifferential}
        )
        assert_representation_allclose(result_c, expected_c)
        assert_differential_allclose(
            result_c.differentials["s"], expected_c.differentials["s"]
        )


@pytest.mark.parametrize("op,args", [(operator.neg, ()), (operator.mul, (10.0,))])
def test_operation_unitspherical_with_rv_fails(op, args):
    rep = UnitSphericalRepresentation(
        0 * u.deg, 0 * u.deg, differentials={"s": RadialDifferential(10 * u.km / u.s)}
    )
    with pytest.raises(ValueError, match="unit key"):
        op(rep, *args)


@pytest.mark.parametrize(
    "rep,dif",
    [
        [
            CartesianRepresentation([1, 2, 3] * u.kpc),
            CartesianDifferential([0.1, 0.2, 0.3] * u.km / u.s),
        ],
        [
            SphericalRepresentation(90 * u.deg, 0.0 * u.deg, 14.0 * u.kpc),
            SphericalDifferential(1.0 * u.deg, 2.0 * u.deg, 0.1 * u.kpc),
        ],
    ],
)
def test_arithmetic_with_differentials_fail(rep, dif):
    rep = rep.with_differentials(dif)

    with pytest.raises(TypeError):
        rep + rep

    with pytest.raises(TypeError):
        rep - rep

    with pytest.raises(TypeError):
        rep * rep

    with pytest.raises(TypeError):
        rep / rep



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# The following few lines skip this module when running tests if matplotlib is
# not available (and will have no impact otherwise)

try:
    import pytest

    pytest.importorskip("matplotlib")
    del pytest
except ImportError:
    pass

from astropy import config as _config

from .coordinate_helpers import CoordinateHelper
from .coordinates_map import CoordinatesMap
from .core import *
from .helpers import *
from .patches import *
from .wcsapi import custom_ucd_coord_meta_mapping


class Conf(_config.ConfigNamespace):
    """
    Configuration parameters for `astropy.visualization.wcsaxes`.
    """

    coordinate_range_samples = _config.ConfigItem(
        50,
        "The number of samples along each image axis when determining "
        "the range of coordinates in a plot.",
    )

    frame_boundary_samples = _config.ConfigItem(
        1000,
        "How many points to sample along the axes when determining tick locations.",
    )

    grid_samples = _config.ConfigItem(
        1000, "How many points to sample along grid lines."
    )

    contour_grid_samples = _config.ConfigItem(
        200, "The grid size to use when drawing a grid using contours"
    )


conf = Conf()



# Licensed under a 3-clause BSD style license - see LICNSE.rst

# This module includes files automatically generated from ply (these end in
# _lextab.py and _parsetab.py). To generate these files, remove them from this
# folder, then build astropy and run the tests in-place:
#
#   python setup.py build_ext --inplace
#   pytest astropy/units
#
# You can then commit the changes to the re-generated _lextab.py and
# _parsetab.py files.

"""
Handles units in `Office of Guest Investigator Programs (OGIP)
FITS files
<https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/general/ogip_93_001/>`__.
"""

import math
import warnings
from fractions import Fraction
from typing import ClassVar, Literal

import numpy as np

from astropy.extern.ply.lex import Lexer
from astropy.units.core import CompositeUnit, UnitBase
from astropy.units.enums import DeprecatedUnitAction
from astropy.units.errors import UnitParserWarning, UnitsWarning
from astropy.units.typing import UnitScale
from astropy.utils import classproperty, parsing
from astropy.utils.parsing import ThreadSafeParser

from .base import Base, _ParsingFormatMixin


class OGIP(Base, _ParsingFormatMixin):
    """
    Support the units in `Office of Guest Investigator Programs (OGIP)
    FITS files
    <https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/general/ogip_93_001/>`__.
    """

    _tokens: ClassVar[tuple[str, ...]] = (
        "DIVISION",
        "OPEN_PAREN",
        "CLOSE_PAREN",
        "WHITESPACE",
        "POWER",
        "STAR",
        "FLOAT",
        "LIT10",
        "INT",
        "UNKNOWN",
        "FUNCNAME",
        "UNIT",
    )

    _deprecated_units: ClassVar[frozenset[str]] = frozenset(("Crab", "mCrab"))

    @classproperty(lazy=True)
    def _units(cls) -> dict[str, UnitBase]:
        from astropy import units as u

        names = {"as": u.attosecond}
        for non_prefixed_unit in [
            "angstrom", "arcmin", "arcsec", "AU", "barn", "bin",
            "byte", "chan", "count", "d", "deg", "erg", "G",
            "h", "lyr", "mag", "min", "photon", "pixel",
            "voxel", "yr",
        ]:  # fmt: skip
            names[non_prefixed_unit] = getattr(u, non_prefixed_unit)

        bases = [
            "A", "C", "cd", "eV", "F", "g", "H", "Hz", "J",
            "Jy", "K", "lm", "lx", "m", "mol", "N", "ohm", "Pa",
            "pc", "rad", "s", "S", "sr", "T", "V", "W", "Wb",
        ]  # fmt: skip
        prefixes = [
            "y", "z", "a", "f", "p", "n", "u", "m", "c", "d",
            "", "da", "h", "k", "M", "G", "T", "P", "E", "Z", "Y",
        ]  # fmt: skip

        for name in (prefix + base for base in bases for prefix in prefixes):
            if name not in names:
                names[name] = getattr(u, name)

        # Create a separate, disconnected unit for the special case of
        # Crab and mCrab, since OGIP doesn't define their quantities.
        names["Crab"] = u.def_unit(["Crab"], prefixes=False, doc="Crab (X-ray flux)")
        names["mCrab"] = u.Unit(10**-3 * names["Crab"])

        return names

    @classproperty(lazy=True)
    def _lexer(cls) -> Lexer:
        tokens = cls._tokens

        t_DIVISION = "[ \t]*/[ \t]*"
        t_OPEN_PAREN = r"\("
        t_CLOSE_PAREN = r"\)"
        t_WHITESPACE = "[ \t]+"
        t_POWER = r"\*\*"
        t_STAR = r"\*"

        # NOTE THE ORDERING OF THESE RULES IS IMPORTANT!!
        # Regular expression rules for simple tokens
        def t_FLOAT(t):
            r"[+-]?((((\d+\.?\d*)|(\.\d+))([eE][+-]?\d+))|(((\d+\.\d*)|(\.\d+))([eE][+-]?\d+)?))"
            t.value = float(t.value)
            return t

        def t_INT(t):
            r"[+-]?\d+"
            t.value = int(t.value)
            return t

        def t_LIT10(t):
            r"10"
            return 10

        def t_UNKNOWN(t):
            r"[Uu][Nn][Kk][Nn][Oo][Ww][Nn]"
            return None

        def t_FUNCNAME(t):
            r"((sqrt)|(ln)|(exp)|(log)|(sin)|(cos)|(tan)|(asin)|(acos)|(atan)|(sinh)|(cosh)|(tanh))(?=\ *\()"
            return t

        def t_UNIT(t):
            r"[a-zA-Z][a-zA-Z_]*"
            t.value = cls._get_unit(t)
            return t

        # Don't ignore whitespace
        t_ignore = ""

        # Error handling rule
        def t_error(t):
            raise ValueError(f"Invalid character at col {t.lexpos}")

        return parsing.lex(lextab="ogip_lextab", package="astropy/units")

    @classproperty(lazy=True)
    def _parser(cls) -> ThreadSafeParser:
        """
        The grammar here is based on the description in the
        `Specification of Physical Units within OGIP FITS files
        <https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/general/ogip_93_001/>`__,
        which is not terribly precise.  The exact grammar is here is
        based on the YACC grammar in the `unity library
        <https://bitbucket.org/nxg/unity/>`_.
        """
        tokens = cls._tokens

        def p_main(p):
            """
            main : UNKNOWN
                 | complete_expression
                 | scale_factor complete_expression
                 | scale_factor WHITESPACE complete_expression
            """
            match p[1:]:
                case (factor, unit) | (factor, _, unit):
                    p[0] = CompositeUnit(factor * unit.scale, unit.bases, unit.powers)
                case _:
                    p[0] = p[1]

        def p_complete_expression(p):
            """
            complete_expression : unit_expression
                                | product_of_units
                                | division_of_units
            """
            # product_of_units is not in unit_expression for performance
            # division_of_units is separate to enforce the correct order of operations
            p[0] = p[1]

        def p_product_of_units(p):
            """
            product_of_units : complete_expression product unit_expression
            """
            p[0] = p[1] * p[3]

        def p_division_of_units(p):
            """
            division_of_units : DIVISION unit_expression
                              | complete_expression DIVISION unit_expression
            """
            match p[1:]:
                case _, unit:
                    p[0] = unit**-1
                case num, _, denom:
                    p[0] = num / denom

        def p_unit_expression(p):
            """
            unit_expression : UNIT
                            | function
                            | UNIT POWER numeric_power
                            | UNIT OPEN_PAREN complete_expression CLOSE_PAREN
                            | OPEN_PAREN complete_expression CLOSE_PAREN
                            | UNIT OPEN_PAREN complete_expression CLOSE_PAREN POWER numeric_power
                            | OPEN_PAREN complete_expression CLOSE_PAREN POWER numeric_power
            """
            bad_multiplication_message = (
                "if '{0}{1}' was meant to be a multiplication, "
                "it should have been written as '{0} {1}'."
            )

            match p[1:]:
                case factor, _, unit, _, _, power:
                    warnings.warn(
                        bad_multiplication_message.format(factor, f"({unit})**{power}"),
                        UnitParserWarning,
                    )
                    p[0] = factor * unit**power
                case (_, unit, _, _, power) | (unit, "**", power):
                    p[0] = unit**power
                case left, _, right, _:
                    warnings.warn(
                        bad_multiplication_message.format(left, f"({right})"),
                        UnitParserWarning,
                    )
                    p[0] = left * right
                case _, unit, _:
                    p[0] = unit
                case _:
                    p[0] = p[1]

        def p_function(p):
            """
            function : FUNCNAME OPEN_PAREN complete_expression CLOSE_PAREN
                     | FUNCNAME OPEN_PAREN complete_expression CLOSE_PAREN POWER numeric_power
            """
            match p[1:]:
                case "sqrt", _, unit, _:
                    p[0] = unit**0.5
                case "sqrt", _, unit, _, _, numeric_power:
                    p[0] = unit ** (0.5 * numeric_power)
                case func, *_:
                    raise ValueError(
                        f"The function '{func}' is valid in OGIP, but not understood "
                        "by astropy.units."
                    )

        def p_scale_factor(p):
            """
            scale_factor : LIT10 POWER numeric_power
                         | LIT10
                         | number
                         | number POWER numeric_power
            """
            if len(p) == 4:
                p[0] = 10 ** p[3]
            else:
                p[0] = p[1]
            # Can't use np.log10 here, because p[0] may be a Python long.
            if math.log10(p[0]) % 1.0 != 0.0:
                warnings.warn(
                    f"'{p[0]}' scale should be a power of 10 in OGIP format",
                    UnitsWarning,
                )

        def p_product(p):
            """
            product : WHITESPACE
                    | STAR
                    | WHITESPACE STAR
                    | WHITESPACE STAR WHITESPACE
                    | STAR WHITESPACE
            """

        def p_numeric_power(p):
            """
            numeric_power : number
                          | OPEN_PAREN number CLOSE_PAREN
                          | OPEN_PAREN INT DIVISION INT CLOSE_PAREN
            """
            if len(p) == 6:
                p[0] = Fraction(int(p[2]), int(p[4]))
            elif len(p) == 4:
                p[0] = p[2]
            else:
                p[0] = p[1]
                if p[1] < 0:
                    warnings.warn(
                        UnitParserWarning(
                            "negative exponents must be enclosed in parenthesis. "
                            f"Expected '**({p[1]})' instead of '**{p[1]}'."
                        )
                    )

        def p_number(p):
            """
            number : INT
                   | FLOAT
            """
            p[0] = p[1]

        def p_error(p):
            raise ValueError()

        return parsing.yacc(tabmodule="ogip_parsetab", package="astropy/units")

    @classmethod
    def parse(cls, s: str, debug: bool = False) -> UnitBase:
        return cls._do_parse(s.strip(), debug)

    @classmethod
    def _format_superscript(cls, number: str) -> str:
        return f"**({number})" if "/" in number else f"**{number}"

    @classmethod
    def to_string(
        cls,
        unit: UnitBase,
        fraction: bool | Literal["inline", "multiline"] = "inline",
        deprecations: DeprecatedUnitAction = DeprecatedUnitAction.WARN,
    ) -> str:
        # Remove units that aren't known to the format
        unit = cls._decompose_to_known_units(unit, deprecations)

        if isinstance(unit, CompositeUnit):
            # Can't use np.log10 here, because p[0] may be a Python long.
            if math.log10(unit.scale) % 1.0 != 0.0:
                warnings.warn(
                    f"'{unit.scale}' scale should be a power of 10 in OGIP format",
                    UnitsWarning,
                )

        return super().to_string(unit, fraction=fraction)

    @classmethod
    def format_exponential_notation(
        cls, val: UnitScale | np.number, format_spec: str = "g"
    ) -> str:
        return format(val, format_spec)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

from astropy.cosmology import Cosmology
from astropy.cosmology._src.core import _COSMOLOGY_CLASSES
from astropy.cosmology._src.io.builtin.table import from_table, to_table
from astropy.table import QTable, Table, vstack

from .base import ToFromDirectTestBase, ToFromTestMixinBase

###############################################################################


class ToFromTableTestMixin(ToFromTestMixinBase):
    """
    Tests for a Cosmology[To/From]Format with ``format="astropy.table"``.
    This class will not be directly called by :mod:`pytest` since its name does
    not begin with ``Test``. To activate the contained tests this class must
    be inherited in a subclass. Subclasses must define a :func:`pytest.fixture`
    ``cosmo`` that returns/yields an instance of a |Cosmology|.
    See ``TestCosmology`` for an example.
    """

    def test_to_table_bad_index(self, from_format, to_format):
        """Test if argument ``index`` is incorrect"""
        tbl = to_format("astropy.table")

        # single-row table and has a non-0/None index
        with pytest.raises(IndexError, match="index 2 out of range"):
            from_format(tbl, index=2, format="astropy.table")

        # string index where doesn't match
        with pytest.raises(KeyError, match="No matches found for key"):
            from_format(tbl, index="row 0", format="astropy.table")

    # -----------------------

    def test_to_table_failed_cls(self, to_format):
        """Test failed table type."""
        with pytest.raises(TypeError, match="'cls' must be"):
            to_format("astropy.table", cls=list)

    @pytest.mark.parametrize("tbl_cls", [QTable, Table])
    def test_to_table_cls(self, to_format, tbl_cls):
        tbl = to_format("astropy.table", cls=tbl_cls)
        assert isinstance(tbl, tbl_cls)  # test type

    # -----------------------

    @pytest.mark.parametrize("in_meta", [True, False])
    def test_to_table_in_meta(self, cosmo_cls, to_format, in_meta):
        """Test where the cosmology class is placed."""
        tbl = to_format("astropy.table", cosmology_in_meta=in_meta)

        # if it's in metadata, it's not a column. And vice versa.
        if in_meta:
            assert tbl.meta["cosmology"] == cosmo_cls.__qualname__
            assert "cosmology" not in tbl.colnames  # not also a column
        else:
            assert tbl["cosmology"][0] == cosmo_cls.__qualname__
            assert "cosmology" not in tbl.meta

    # -----------------------

    def test_to_table(self, cosmo_cls, cosmo, to_format):
        """Test cosmology -> astropy.table."""
        tbl = to_format("astropy.table")

        # Test properties of Table.
        assert isinstance(tbl, QTable)
        assert tbl.meta["cosmology"] == cosmo_cls.__qualname__
        assert tbl["name"] == cosmo.name
        assert tbl.indices  # indexed

        # Test each Parameter column has expected information.
        for n, P in cosmo_cls.parameters.items():
            col = tbl[n]  # Column

            # Compare the two
            assert col.info.name == P.name
            assert col.info.description == P.__doc__
            assert col.info.meta == (cosmo.meta.get(n) or {})

    # -----------------------

    def test_from_not_table(self, cosmo, from_format):
        """Test not passing a Table to the Table parser."""
        with pytest.raises((TypeError, ValueError)):
            from_format("NOT A TABLE", format="astropy.table")

    def test_tofrom_table_instance(self, cosmo_cls, cosmo, from_format, to_format):
        """Test cosmology -> astropy.table -> cosmology."""
        tbl = to_format("astropy.table")

        # add information
        tbl["mismatching"] = "will error"

        # tests are different if the last argument is a **kwarg
        if cosmo._init_has_kwargs:
            got = from_format(tbl, format="astropy.table")

            assert got.__class__ is cosmo_cls
            assert got.name == cosmo.name
            assert "mismatching" not in got.meta

            return  # don't continue testing

        # read with mismatching parameters errors
        with pytest.raises(TypeError, match="there are unused parameters"):
            from_format(tbl, format="astropy.table")

        # unless mismatched are moved to meta
        got = from_format(tbl, format="astropy.table", move_to_meta=True)
        assert got == cosmo
        assert got.meta["mismatching"] == "will error"

        # it won't error if everything matches up
        tbl.remove_column("mismatching")
        got = from_format(tbl, format="astropy.table")
        assert got == cosmo

        # and it will also work if the cosmology is a class
        # Note this is not the default output of ``to_format``.
        tbl.meta["cosmology"] = _COSMOLOGY_CLASSES[tbl.meta["cosmology"]]
        got = from_format(tbl, format="astropy.table")
        assert got == cosmo

        # also it auto-identifies 'format'
        got = from_format(tbl)
        assert got == cosmo

    def test_fromformat_table_subclass_partial_info(
        self, cosmo_cls, cosmo, from_format, to_format
    ):
        """
        Test writing from an instance and reading from that class.
        This works with missing information.
        """
        # test to_format
        tbl = to_format("astropy.table")
        assert isinstance(tbl, QTable)

        # partial information
        tbl.meta.pop("cosmology", None)
        del tbl["Tcmb0"]

        # read with the same class that wrote fills in the missing info with
        # the default value
        got = cosmo_cls.from_format(tbl, format="astropy.table")
        got2 = from_format(tbl, format="astropy.table", cosmology=cosmo_cls)
        got3 = from_format(
            tbl, format="astropy.table", cosmology=cosmo_cls.__qualname__
        )

        assert (got == got2) and (got2 == got3)  # internal consistency

        # not equal, because Tcmb0 is changed, which also changes m_nu
        assert got != cosmo
        assert got.Tcmb0 == cosmo_cls.parameters["Tcmb0"].default
        assert got.clone(name=cosmo.name, Tcmb0=cosmo.Tcmb0, m_nu=cosmo.m_nu) == cosmo
        # but the metadata is the same
        assert got.meta == cosmo.meta

    @pytest.mark.parametrize("add_index", [True, False])
    def test_tofrom_table_mutlirow(self, cosmo_cls, cosmo, from_format, add_index):
        """Test if table has multiple rows."""
        # ------------
        # To Table

        cosmo1 = cosmo.clone(name="row 0")
        cosmo2 = cosmo.clone(name="row 2")
        tbl = vstack(
            [c.to_format("astropy.table") for c in (cosmo1, cosmo, cosmo2)],
            metadata_conflicts="silent",
        )

        assert isinstance(tbl, QTable)
        assert tbl.meta["cosmology"] == cosmo_cls.__qualname__
        assert tbl[1]["name"] == cosmo.name

        # whether to add an index. `from_format` can work with or without.
        if add_index:
            tbl.add_index("name", unique=True)

        # ------------
        # From Table

        # it will error on a multi-row table
        with pytest.raises(ValueError, match="need to select a specific row"):
            from_format(tbl, format="astropy.table")

        # unless the index argument is provided
        got = from_format(tbl, index=1, format="astropy.table")
        assert got == cosmo

        # the index can be a string
        got = from_format(tbl, index=cosmo.name, format="astropy.table")
        assert got == cosmo

        # when there's more than one cosmology found
        tbls = vstack([tbl, tbl], metadata_conflicts="silent")
        with pytest.raises(ValueError, match="more than one"):
            from_format(tbls, index=cosmo.name, format="astropy.table")

    def test_tofrom_table_rename(self, cosmo, to_format, from_format):
        """Test renaming columns in row."""
        rename = {"name": "cosmo_name"}
        table = to_format("astropy.table", rename=rename)

        assert "name" not in table.colnames
        assert "cosmo_name" in table.colnames

        # Error if just reading
        with pytest.raises(TypeError, match="there are unused parameters"):
            from_format(table)

        # Roundtrip
        inv_rename = {v: k for k, v in rename.items()}
        got = from_format(table, rename=inv_rename)
        assert got == cosmo

    def test_from_table_renamed_index_column(self, cosmo, to_format, from_format):
        """Test reading from a table with a renamed index column."""
        cosmo1 = cosmo.clone(name="row 0")
        cosmo2 = cosmo.clone(name="row 2")
        tbl = vstack(
            [c.to_format("astropy.table") for c in (cosmo1, cosmo, cosmo2)],
            metadata_conflicts="silent",
        )
        tbl.rename_column("name", "cosmo_name")

        inv_rename = {"cosmo_name": "name"}
        newcosmo = from_format(
            tbl, index="row 0", rename=inv_rename, format="astropy.table"
        )
        assert newcosmo == cosmo1

    @pytest.mark.parametrize("format", [True, False, None, "astropy.table"])
    def test_is_equivalent_to_table(self, cosmo, to_format, format):
        """Test :meth:`astropy.cosmology.Cosmology.is_equivalent`.

        This test checks that Cosmology equivalency can be extended to any
        Python object that can be converted to a Cosmology -- in this case
        a |Table|.
        """
        obj = to_format("astropy.table")
        assert not isinstance(obj, Cosmology)

        is_equiv = cosmo.is_equivalent(obj, format=format)
        assert is_equiv is (format is not False)


class TestToFromTable(ToFromDirectTestBase, ToFromTableTestMixin):
    """Directly test ``to/from_table``."""

    def setup_class(self):
        self.functions = {"to": to_table, "from": from_table}



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import inspect
import sys

import pytest

from astropy import cosmology
from astropy.cosmology import Cosmology, w0wzCDM
from astropy.cosmology._src.tests.io import (
    test_cosmology,
    test_ecsv,
    test_html,
    test_json,
    test_latex,
    test_mapping,
    test_model,
    test_row,
    test_table,
    test_yaml,
)
from astropy.cosmology.io import readwrite_registry
from astropy.table import QTable, Row
from astropy.utils.compat.optional_deps import HAS_BS4

###############################################################################
# SETUP

cosmo_instances = cosmology.realizations.available

# Collect the registered read/write formats.
#   (format, supports_metadata, has_all_required_dependencies)
readwrite_formats = [
    ("ascii.ecsv", True, True),
    ("ascii.html", False, HAS_BS4),
    ("ascii.latex", False, True),
    ("json", True, True),
    ("latex", False, True),
]


# Collect all the registered to/from formats. Unfortunately this is NOT
# automatic since the output format class is not stored on the registry.
#                 (format, data type)
tofrom_formats = [
    ("mapping", dict),
    ("yaml", str),
    ("astropy.cosmology", Cosmology),
    ("astropy.row", Row),
    ("astropy.table", QTable),
]


###############################################################################


class ReadWriteTestMixin(
    test_ecsv.ReadWriteECSVTestMixin,
    test_html.ReadWriteHTMLTestMixin,
    test_json.ReadWriteJSONTestMixin,
    test_latex.WriteLATEXTestMixin,
):
    """
    Tests for a CosmologyRead/Write on a |Cosmology|.
    This class will not be directly called by :mod:`pytest` since its name does
    not begin with ``Test``. To activate the contained tests this class must
    be inherited in a subclass. Subclasses must define a :func:`pytest.fixture`
    ``cosmo`` that returns/yields an instance of a |Cosmology|.
    See ``TestReadWriteCosmology`` or ``TestCosmology`` for examples.
    """

    @pytest.mark.parametrize("format, metaio, has_deps", readwrite_formats)
    def test_readwrite_complete_info(self, cosmo, tmp_path, format, metaio, has_deps):
        """
        Test writing from an instance and reading from the base class.
        This requires full information.
        The round-tripped metadata can be in a different order, so the
        OrderedDict must be converted to a dict before testing equality.
        """
        if not has_deps:
            pytest.skip("missing a dependency")
        if (format, Cosmology) not in readwrite_registry._readers:
            pytest.xfail(f"no read method is registered for format {format!r}")

        fname = tmp_path / f"{cosmo.name}.{format}"
        cosmo.write(fname, format=format)

        # Also test kwarg "overwrite"
        assert fname.is_file()
        with pytest.raises(IOError):
            cosmo.write(fname, format=format, overwrite=False)

        assert fname.exists()  # overwrite file existing file
        cosmo.write(fname, format=format, overwrite=True)

        # Read back
        got = Cosmology.read(fname, format=format)

        assert got == cosmo
        assert (not metaio) ^ (dict(got.meta) == dict(cosmo.meta))

    @pytest.mark.parametrize("format, metaio, has_deps", readwrite_formats)
    def test_readwrite_from_subclass_complete_info(
        self, cosmo_cls, cosmo, tmp_path, format, metaio, has_deps
    ):
        """
        Test writing from an instance and reading from that class, when there's
        full information saved.
        """
        if not has_deps:
            pytest.skip("missing a dependency")
        if (format, Cosmology) not in readwrite_registry._readers:
            pytest.xfail(f"no read method is registered for format {format!r}")

        fname = str(tmp_path / f"{cosmo.name}.{format}")
        cosmo.write(fname, format=format)

        # read with the same class that wrote.
        got = cosmo_cls.read(fname, format=format)
        assert got == cosmo
        assert (not metaio) ^ (dict(got.meta) == dict(cosmo.meta))

        # this should be equivalent to
        got = Cosmology.read(fname, format=format, cosmology=cosmo_cls)
        assert got == cosmo
        assert (not metaio) ^ (dict(got.meta) == dict(cosmo.meta))

        # and also
        got = Cosmology.read(fname, format=format, cosmology=cosmo_cls.__qualname__)
        assert got == cosmo
        assert (not metaio) ^ (dict(got.meta) == dict(cosmo.meta))


class TestCosmologyReadWrite(ReadWriteTestMixin):
    """Test the classes CosmologyRead/Write."""

    @pytest.fixture(scope="class", params=cosmo_instances)
    def cosmo(self, request):
        return getattr(cosmology.realizations, request.param)

    @pytest.fixture(scope="class")
    def cosmo_cls(self, cosmo):
        return cosmo.__class__

    # ==============================================================

    @pytest.mark.parametrize("format, _, has_deps", readwrite_formats)
    def test_write_methods_have_explicit_kwarg_overwrite(self, format, _, has_deps):
        if not has_deps:
            pytest.skip("missing a dependency")
        if (format, Cosmology) not in readwrite_registry._readers:
            pytest.xfail(f"no read method is registered for format {format!r}")

        writer = readwrite_registry.get_writer(format, Cosmology)
        # test in signature
        sig = inspect.signature(writer)
        assert "overwrite" in sig.parameters

        # also in docstring
        if not sys.flags.optimize:
            assert "overwrite : bool" in writer.__doc__

    @pytest.mark.parametrize("format, _, has_deps", readwrite_formats)
    def test_readwrite_reader_class_mismatch(
        self, cosmo, tmp_path, format, _, has_deps
    ):
        """Test when the reader class doesn't match the file."""
        if not has_deps:
            pytest.skip("missing a dependency")
        if (format, Cosmology) not in readwrite_registry._readers:
            pytest.xfail(f"no read method is registered for format {format!r}")

        fname = tmp_path / f"{cosmo.name}.{format}"
        cosmo.write(fname, format=format)

        # class mismatch
        # when reading directly
        with pytest.raises(TypeError, match="missing 1 required"):
            w0wzCDM.read(fname, format=format)

        with pytest.raises(TypeError, match="missing 1 required"):
            Cosmology.read(fname, format=format, cosmology=w0wzCDM)

        # when specifying the class
        with pytest.raises(ValueError, match="`cosmology` must be either"):
            w0wzCDM.read(fname, format=format, cosmology="FlatLambdaCDM")


###############################################################################
# To/From_Format Tests


class ToFromFormatTestMixin(
    test_cosmology.ToFromCosmologyTestMixin,
    test_mapping.ToFromMappingTestMixin,
    test_model.ToFromModelTestMixin,
    test_row.ToFromRowTestMixin,
    test_table.ToFromTableTestMixin,
    test_yaml.ToFromYAMLTestMixin,
):
    """
    Tests for a Cosmology[To/From]Format on a |Cosmology|.
    This class will not be directly called by :mod:`pytest` since its name does
    not begin with ``Test``. To activate the contained tests this class must
    be inherited in a subclass. Subclasses must define a :func:`pytest.fixture`
    ``cosmo`` that returns/yields an instance of a |Cosmology|.
    See ``TestCosmology`` for an example.
    """

    @pytest.mark.parametrize("format, totype", tofrom_formats)
    def test_tofromformat_complete_info(
        self, cosmo, format, totype, xfail_if_not_registered_with_yaml
    ):
        """Read tests happen later."""
        # test to_format
        obj = cosmo.to_format(format)
        assert isinstance(obj, totype)

        # test from_format
        got = Cosmology.from_format(obj, format=format)

        # Test autodetect, if enabled
        if self.can_autodentify(format):
            got2 = Cosmology.from_format(obj)
            assert got2 == got  # internal consistency

        assert got == cosmo  # external consistency
        assert got.meta == cosmo.meta

    @pytest.mark.parametrize("format, totype", tofrom_formats)
    def test_fromformat_subclass_complete_info(
        self, cosmo_cls, cosmo, format, totype, xfail_if_not_registered_with_yaml
    ):
        """
        Test transforming an instance and parsing from that class, when there's
        full information available.
        Partial information tests are handled in the Mixin super classes.
        """
        # test to_format
        obj = cosmo.to_format(format)
        assert isinstance(obj, totype)

        # read with the same class that wrote.
        got = cosmo_cls.from_format(obj, format=format)

        if self.can_autodentify(format):
            got2 = Cosmology.from_format(obj)  # and autodetect
            assert got2 == got  # internal consistency

        assert got == cosmo  # external consistency
        assert got.meta == cosmo.meta

        # this should be equivalent to
        got = Cosmology.from_format(obj, format=format, cosmology=cosmo_cls)
        assert got == cosmo
        assert got.meta == cosmo.meta

        # and also
        got = Cosmology.from_format(
            obj, format=format, cosmology=cosmo_cls.__qualname__
        )
        assert got == cosmo
        assert got.meta == cosmo.meta


class TestCosmologyToFromFormat(ToFromFormatTestMixin):
    """Test Cosmology[To/From]Format classes."""

    @pytest.fixture(scope="class", params=cosmo_instances)
    def cosmo(self, request):
        return getattr(cosmology.realizations, request.param)

    @pytest.fixture(scope="class")
    def cosmo_cls(self, cosmo):
        return cosmo.__class__

    # ==============================================================

    @pytest.mark.parametrize("format_type", tofrom_formats)
    def test_fromformat_class_mismatch(self, cosmo, format_type):
        format, totype = format_type

        # test to_format
        obj = cosmo.to_format(format)
        assert isinstance(obj, totype)

        # class mismatch
        with pytest.raises(TypeError):
            w0wzCDM.from_format(obj, format=format)

        with pytest.raises(TypeError):
            Cosmology.from_format(obj, format=format, cosmology=w0wzCDM)

        # when specifying the class
        with pytest.raises(ValueError, match="`cosmology` must be either"):
            w0wzCDM.from_format(obj, format=format, cosmology="FlatLambdaCDM")



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy.stats.histogram import calculate_bin_edges

__all__ = ["hist"]


def hist(x, bins=10, ax=None, max_bins=1e5, **kwargs):
    """Enhanced histogram function.

    This is a histogram function that enables the use of more sophisticated
    algorithms for determining bins.  Aside from the ``bins`` argument allowing
    a string specified how bins are computed, the parameters are the same
    as matplotlib.pyplot.hist().

    This function was ported from astroML: https://www.astroml.org/

    Parameters
    ----------
    x : array-like
        array of data to be histogrammed

    bins : int, list, or str, optional
        If bins is a string, then it must be one of:

        - 'blocks' : use bayesian blocks for dynamic bin widths

        - 'knuth' : use Knuth's rule to determine bins

        - 'scott' : use Scott's rule to determine bins

        - 'freedman' : use the Freedman-Diaconis rule to determine bins

    ax : `~matplotlib.axes.Axes` instance, optional
        Specify the Axes on which to draw the histogram. If not specified,
        then the current active axes will be used.

    max_bins : int, optional
        Maximum number of bins allowed. With more than a few thousand bins
        the performance of matplotlib will not be great. If the number of
        bins is large *and* the number of input data points is large then
        the it will take a very long time to compute the histogram.

    **kwargs :
        other keyword arguments are described in ``plt.hist()``.

    Notes
    -----
    Return values are the same as for ``plt.hist()``

    See Also
    --------
    astropy.stats.histogram
    """
    # Note that we only calculate the bin edges...matplotlib will calculate
    # the actual histogram.
    range = kwargs.get("range")
    weights = kwargs.get("weights")
    bins = calculate_bin_edges(x, bins, range=range, weights=weights)

    if len(bins) > max_bins:
        raise ValueError(
            "Histogram has too many bins: "
            f"{len(bins)}. Use max_bins to increase the number "
            "of allowed bins or range to restrict "
            "the histogram range."
        )

    if ax is None:
        # optional dependency; only import if strictly needed.
        import matplotlib.pyplot as plt

        ax = plt.gca()

    return ax.hist(x, bins, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from collections.abc import Callable
from typing import TypeVar

import numpy as np
from numpy.typing import NDArray

__all__ = ["jackknife_resampling", "jackknife_stats"]
__doctest_requires__ = {"jackknife_stats": ["scipy"]}

DT = TypeVar("DT", bound=np.generic)


def jackknife_resampling(data: NDArray[DT]) -> NDArray[DT]:
    """Performs jackknife resampling on numpy arrays.

    Jackknife resampling is a technique to generate 'n' deterministic samples
    of size 'n-1' from a measured sample of size 'n'. Basically, the i-th
    sample, (1<=i<=n), is generated by means of removing the i-th measurement
    of the original sample. Like the bootstrap resampling, this statistical
    technique finds applications in estimating variance, bias, and confidence
    intervals.

    Parameters
    ----------
    data : ndarray
        Original sample (1-D array) from which the jackknife resamples will be
        generated.

    Returns
    -------
    resamples : ndarray
        The i-th row is the i-th jackknife sample, i.e., the original sample
        with the i-th measurement deleted.

    References
    ----------
    .. [1] McIntosh, Avery. "The Jackknife Estimation Method".
        <https://arxiv.org/abs/1606.00497>

    .. [2] Efron, Bradley. "The Jackknife, the Bootstrap, and other
        Resampling Plans". Technical Report No. 63, Division of Biostatistics,
        Stanford University, December, 1980.

    .. [3] Jackknife resampling <https://en.wikipedia.org/wiki/Jackknife_resampling>
    """
    n = data.shape[0]
    if n <= 0:
        raise ValueError("data must contain at least one measurement.")

    resamples = np.empty([n, n - 1])

    for i in range(n):
        resamples[i] = np.delete(data, i)

    return resamples


def jackknife_stats(
    data: NDArray,
    statistic: Callable,
    confidence_level: float = 0.95,
) -> tuple[float | NDArray, float | NDArray, float | NDArray, NDArray]:
    """Performs jackknife estimation on the basis of jackknife resamples.

    This function requires `SciPy <https://www.scipy.org/>`_ to be installed.

    Parameters
    ----------
    data : ndarray
        Original sample (1-D array).
    statistic : function
        Any function (or vector of functions) on the basis of the measured
        data, e.g, sample mean, sample variance, etc. The jackknife estimate of
        this statistic will be returned.
    confidence_level : float, optional
        Confidence level for the confidence interval of the Jackknife estimate.
        Must be a real-valued number in (0,1). Default value is 0.95.

    Returns
    -------
    estimate : float or `~numpy.ndarray`
        The i-th element is the bias-corrected "jackknifed" estimate.

    bias : float or `~numpy.ndarray`
        The i-th element is the jackknife bias.

    std_err : float or `~numpy.ndarray`
        The i-th element is the jackknife standard error.

    conf_interval : ndarray
        If ``statistic`` is single-valued, the first and second elements are
        the lower and upper bounds, respectively. If ``statistic`` is
        vector-valued, each column corresponds to the confidence interval for
        each component of ``statistic``. The first and second rows contain the
        lower and upper bounds, respectively.

    Examples
    --------
    1. Obtain Jackknife resamples:

    >>> import numpy as np
    >>> from astropy.stats import jackknife_resampling
    >>> from astropy.stats import jackknife_stats
    >>> data = np.array([1,2,3,4,5,6,7,8,9,0])
    >>> resamples = jackknife_resampling(data)
    >>> resamples
    array([[2., 3., 4., 5., 6., 7., 8., 9., 0.],
           [1., 3., 4., 5., 6., 7., 8., 9., 0.],
           [1., 2., 4., 5., 6., 7., 8., 9., 0.],
           [1., 2., 3., 5., 6., 7., 8., 9., 0.],
           [1., 2., 3., 4., 6., 7., 8., 9., 0.],
           [1., 2., 3., 4., 5., 7., 8., 9., 0.],
           [1., 2., 3., 4., 5., 6., 8., 9., 0.],
           [1., 2., 3., 4., 5., 6., 7., 9., 0.],
           [1., 2., 3., 4., 5., 6., 7., 8., 0.],
           [1., 2., 3., 4., 5., 6., 7., 8., 9.]])
    >>> resamples.shape
    (10, 9)

    2. Obtain Jackknife estimate for the mean, its bias, its standard error,
    and its 95% confidence interval:

    >>> test_statistic = np.mean
    >>> estimate, bias, stderr, conf_interval = jackknife_stats(
    ...     data, test_statistic, 0.95)
    >>> estimate
    np.float64(4.5)
    >>> bias
    np.float64(0.0)
    >>> stderr  # doctest: +FLOAT_CMP
    np.float64(0.95742710775633832)
    >>> conf_interval
    array([2.62347735,  6.37652265])

    3. Example for two estimates

    >>> test_statistic = lambda x: (np.mean(x), np.var(x))
    >>> estimate, bias, stderr, conf_interval = jackknife_stats(
    ...     data, test_statistic, 0.95)
    >>> estimate
    array([4.5       ,  9.16666667])
    >>> bias
    array([ 0.        , -0.91666667])
    >>> stderr
    array([0.95742711,  2.69124476])
    >>> conf_interval
    array([[ 2.62347735,   3.89192387],
           [ 6.37652265,  14.44140947]])

    IMPORTANT: Note that confidence intervals are given as columns
    """
    # jackknife confidence interval
    if not (0 < confidence_level < 1):
        raise ValueError("confidence level must be in (0, 1).")

    # make sure original data is proper
    n = data.shape[0]
    if n <= 0:
        raise ValueError("data must contain at least one measurement.")

    # Only import scipy if inputs are valid
    from scipy.special import erfinv

    resamples = jackknife_resampling(data)

    stat_data = statistic(data)
    jack_stat = np.apply_along_axis(statistic, 1, resamples)
    mean_jack_stat = np.mean(jack_stat, axis=0)

    # jackknife bias
    bias = (n - 1) * (mean_jack_stat - stat_data)

    # jackknife standard error
    std_err = np.sqrt(
        (n - 1)
        * np.mean((jack_stat - mean_jack_stat) * (jack_stat - mean_jack_stat), axis=0)
    )

    # bias-corrected "jackknifed estimate"
    estimate = stat_data - bias

    z_score = np.sqrt(2.0) * erfinv(confidence_level)
    conf_interval = estimate + z_score * np.array((-std_err, std_err))

    return estimate, bias, std_err, conf_interval



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import re

import numpy as np
import pytest

from astropy import __version__ as version
from astropy.io import fits
from astropy.io.fits.scripts import fitscheck
from astropy.utils.exceptions import AstropyUserWarning

from .conftest import FitsTestCase


class TestFitscheck(FitsTestCase):
    def test_help(self):
        with pytest.raises(SystemExit) as e:
            fitscheck.main(["-h"])
        assert e.value.code == 0

    def test_version(self, capsys):
        with pytest.raises(SystemExit) as e:
            fitscheck.main(["--version"])
            out = capsys.readouterr()[0]
            assert out == f"fitscheck {version}"
        assert e.value.code == 0

    def test_missing_file(self, capsys):
        assert fitscheck.main(["missing.fits"]) == 1
        stdout, stderr = capsys.readouterr()
        assert "No such file or directory" in stderr

    def test_valid_file(self, capsys):
        testfile = self.data("checksum.fits")

        assert fitscheck.main([testfile]) == 0
        assert fitscheck.main([testfile, "--compliance"]) == 0

        assert fitscheck.main([testfile, "-v"]) == 0
        stdout, stderr = capsys.readouterr()
        assert "OK" in stderr

    def test_remove_checksums(self, capsys):
        testfile = self.copy_file("checksum.fits")
        assert fitscheck.main([testfile, "--checksum", "remove"]) == 1
        assert fitscheck.main([testfile]) == 1
        stdout, stderr = capsys.readouterr()
        assert "MISSING" in stderr

    def test_no_checksums(self, capsys):
        testfile = self.data("arange.fits")

        assert fitscheck.main([testfile]) == 1
        stdout, stderr = capsys.readouterr()
        assert "Checksum not found" in stderr

        assert fitscheck.main([testfile, "--ignore-missing"]) == 0
        stdout, stderr = capsys.readouterr()
        assert stderr == ""

    def test_overwrite_invalid(self, caplog):
        """
        Tests that invalid checksum or datasum are overwritten when the file is
        saved.
        """
        reffile = self.temp("ref.fits")
        with fits.open(self.data("tb.fits")) as hdul:
            hdul.writeto(reffile, checksum=True)

        # replace checksums with wrong ones
        testfile = self.temp("test.fits")
        with fits.open(self.data("tb.fits")) as hdul:
            hdul[0].header["DATASUM"] = "1       "
            hdul[0].header["CHECKSUM"] = "8UgqATfo7TfoATfo"
            hdul[1].header["DATASUM"] = "2349680925"
            hdul[1].header["CHECKSUM"] = "11daD8bX98baA8bU"
            hdul.writeto(testfile)

        assert fitscheck.main([testfile]) == 1
        for i in range(2):
            assert re.match(
                r"BAD.*Checksum verification failed for HDU",
                caplog.records[i * 2].message,
            )
            assert re.match(
                r"BAD.*Datasum verification failed for HDU",
                caplog.records[i * 2 + 1].message,
            )
        assert re.match(r"4 errors", caplog.records[4].message)
        caplog.clear()

        with pytest.warns(AstropyUserWarning):
            assert fitscheck.main([testfile, "--write", "--force"]) == 1
        assert re.match(
            r"BAD.*Checksum verification failed for HDU", caplog.records[0].message
        )
        caplog.clear()

        # check that the file was fixed
        assert fitscheck.main([testfile]) == 0

    def test_missing_invalid(self, caplog):
        """
        from https://github.com/astropy/astropy/issues/16551
        written by Zach Claytor
        """
        # Test with a file where primary HDU has no checksum/datasum and image
        # extension contains valid checksum/datasum
        testfile = self.temp("test.fits")
        hdul = fits.HDUList([fits.PrimaryHDU()])
        ext = fits.ImageHDU(data=np.arange(10))
        ext.header["THINGY"] = 123
        ext.add_datasum()
        ext.add_checksum()
        hdul.append(ext)
        hdul.writeto(testfile)

        assert fitscheck.main([testfile]) == 1

        assert re.match(
            r"MISSING '.*test\.fits' .. Checksum not found in HDU #0",
            caplog.records[0].message,
        )
        assert re.match(
            r"MISSING '.*test\.fits' .. Datasum not found in HDU #0",
            caplog.records[1].message,
        )
        assert caplog.records[2].message == "2 errors"
        caplog.clear()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Critical density component."""

__all__ = ["CriticalDensity"]

from collections.abc import Callable
from typing import Any

from numpy.typing import ArrayLike, NDArray

from astropy.cosmology._src.utils import deprecated_keywords
from astropy.units import Quantity


class CriticalDensity:
    """The object has attributes and methods for the critical density."""

    critical_density0: Quantity
    """Critical density at redshift 0."""

    efunc: Callable[[Any], NDArray[Any]]

    @deprecated_keywords("z", since="7.0")
    def critical_density(self, z: Quantity | ArrayLike) -> Quantity:
        """Critical density in grams per cubic cm at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        rho : Quantity ['mass density']
            Critical density at each input redshift.
        """
        return self.critical_density0 * self.efunc(z) ** 2



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Contains the transformation functions for getting to/from ITRS, TEME, GCRS, and CIRS.
These are distinct from the ICRS and AltAz functions because they are just
rotations without aberration corrections or offsets.
"""

import erfa
import numpy as np

from astropy.coordinates.baseframe import frame_transform_graph
from astropy.coordinates.matrix_utilities import matrix_transpose
from astropy.coordinates.transformations import FunctionTransformWithFiniteDifference

from .cirs import CIRS
from .equatorial import TEME, TETE
from .gcrs import GCRS, PrecessedGeocentric
from .icrs import ICRS
from .itrs import ITRS
from .utils import get_jd12, get_polar_motion

# # first define helper functions


def teme_to_itrs_mat(time):
    # Sidereal time, rotates from ITRS to mean equinox
    # Use 1982 model for consistency with Vallado et al (2006)
    # https://celestrak.org/publications/aiaa/2006-6753/AIAA-2006-6753.pdf
    gst = erfa.gmst82(*get_jd12(time, "ut1"))

    # Polar Motion
    # Do not include TIO locator s' because it is not used in Vallado 2006
    xp, yp = get_polar_motion(time)
    pmmat = erfa.pom00(xp, yp, 0)

    # rotation matrix
    # c2tcio expects a GCRS->CIRS matrix as it's first argument.
    # Here, we just set that to an I-matrix, because we're already
    # in TEME and the difference between TEME and CIRS is just the
    # rotation by the sidereal time rather than the Earth Rotation Angle
    return erfa.c2tcio(np.eye(3), gst, pmmat)


def gcrs_to_cirs_mat(time):
    # celestial-to-intermediate matrix
    return erfa.c2i06a(*get_jd12(time, "tt"))


def cirs_to_itrs_mat(time):
    # compute the polar motion p-matrix
    xp, yp = get_polar_motion(time)
    sp = erfa.sp00(*get_jd12(time, "tt"))
    pmmat = erfa.pom00(xp, yp, sp)

    # now determine the Earth Rotation Angle for the input obstime
    # era00 accepts UT1, so we convert if need be
    era = erfa.era00(*get_jd12(time, "ut1"))

    # c2tcio expects a GCRS->CIRS matrix, but we just set that to an I-matrix
    # because we're already in CIRS
    return erfa.c2tcio(np.eye(3), era, pmmat)


def tete_to_itrs_mat(time, rbpn=None):
    """Compute the polar motion p-matrix at the given time.

    If the nutation-precession matrix is already known, it should be passed in,
    as this is by far the most expensive calculation.
    """
    xp, yp = get_polar_motion(time)
    sp = erfa.sp00(*get_jd12(time, "tt"))
    pmmat = erfa.pom00(xp, yp, sp)

    # now determine the greenwich apparent sidereal time for the input obstime
    # we use the 2006A model for consistency with RBPN matrix use in GCRS <-> TETE
    ujd1, ujd2 = get_jd12(time, "ut1")
    jd1, jd2 = get_jd12(time, "tt")
    if rbpn is None:
        # erfa.gst06a calls pnm06a to calculate rbpn and then gst06. Use it in
        # favour of getting rbpn with erfa.pnm06a to avoid a possibly large array.
        gast = erfa.gst06a(ujd1, ujd2, jd1, jd2)
    else:
        gast = erfa.gst06(ujd1, ujd2, jd1, jd2, rbpn)

    # c2tcio expects a GCRS->CIRS matrix, but we just set that to an I-matrix
    # because we're already in CIRS equivalent frame
    return erfa.c2tcio(np.eye(3), gast, pmmat)


def gcrs_precession_mat(equinox):
    gamb, phib, psib, epsa = erfa.pfw06(*get_jd12(equinox, "tt"))
    return erfa.fw2m(gamb, phib, psib, epsa)


def get_location_gcrs(location, obstime, ref_to_itrs, gcrs_to_ref):
    """Create a GCRS frame at the location and obstime.

    The reference frame z axis must point to the Celestial Intermediate Pole
    (as is the case for CIRS and TETE).

    This function is here to avoid location.get_gcrs(obstime), which would
    recalculate matrices that are already available below (and return a GCRS
    coordinate, rather than a frame with obsgeoloc and obsgeovel).  Instead,
    it uses the private method that allows passing in the matrices.

    """
    obsgeoloc, obsgeovel = location._get_gcrs_posvel(obstime, ref_to_itrs, gcrs_to_ref)
    return GCRS(obstime=obstime, obsgeoloc=obsgeoloc, obsgeovel=obsgeovel)


# now the actual transforms


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, GCRS, TETE)
def gcrs_to_tete(gcrs_coo, tete_frame):
    # Classical NPB matrix, IAU 2006/2000A
    # (same as in builtin_frames.utils.get_cip).
    rbpn = erfa.pnm06a(*get_jd12(tete_frame.obstime, "tt"))
    # Get GCRS coordinates for the target observer location and time.
    loc_gcrs = get_location_gcrs(
        tete_frame.location,
        tete_frame.obstime,
        tete_to_itrs_mat(tete_frame.obstime, rbpn=rbpn),
        rbpn,
    )
    gcrs_coo2 = gcrs_coo.transform_to(loc_gcrs)
    # Now we are relative to the correct observer, do the transform to TETE.
    # These rotations are defined at the geocenter, but can be applied to
    # topocentric positions as well, assuming rigid Earth. See p57 of
    # https://www.usno.navy.mil/USNO/astronomical-applications/publications/Circular_179.pdf
    crepr = gcrs_coo2.cartesian.transform(rbpn)
    return tete_frame.realize_frame(crepr)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, TETE, GCRS)
def tete_to_gcrs(tete_coo, gcrs_frame):
    # Compute the pn matrix, and then multiply by its transpose.
    rbpn = erfa.pnm06a(*get_jd12(tete_coo.obstime, "tt"))
    newrepr = tete_coo.cartesian.transform(matrix_transpose(rbpn))
    # We now have a GCRS vector for the input location and obstime.
    # Turn it into a GCRS frame instance.
    loc_gcrs = get_location_gcrs(
        tete_coo.location,
        tete_coo.obstime,
        tete_to_itrs_mat(tete_coo.obstime, rbpn=rbpn),
        rbpn,
    )
    gcrs = loc_gcrs.realize_frame(newrepr)
    # Finally, do any needed offsets (no-op if same obstime and location)
    return gcrs.transform_to(gcrs_frame)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, TETE, ITRS)
def tete_to_itrs(tete_coo, itrs_frame):
    # first get us to TETE at the target obstime, and location (no-op if same)
    tete_coo2 = tete_coo.transform_to(
        TETE(obstime=itrs_frame.obstime, location=itrs_frame.location)
    )

    # now get the pmatrix
    pmat = tete_to_itrs_mat(itrs_frame.obstime)
    crepr = tete_coo2.cartesian.transform(pmat)
    return itrs_frame.realize_frame(crepr)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, ITRS, TETE)
def itrs_to_tete(itrs_coo, tete_frame):
    # compute the pmatrix, and then multiply by its transpose
    pmat = tete_to_itrs_mat(itrs_coo.obstime)
    newrepr = itrs_coo.cartesian.transform(matrix_transpose(pmat))
    tete = TETE(newrepr, obstime=itrs_coo.obstime, location=itrs_coo.location)

    # now do any needed offsets (no-op if same obstime and location)
    return tete.transform_to(tete_frame)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, GCRS, CIRS)
def gcrs_to_cirs(gcrs_coo, cirs_frame):
    # first get the pmatrix
    pmat = gcrs_to_cirs_mat(cirs_frame.obstime)
    # Get GCRS coordinates for the target observer location and time.
    loc_gcrs = get_location_gcrs(
        cirs_frame.location,
        cirs_frame.obstime,
        cirs_to_itrs_mat(cirs_frame.obstime),
        pmat,
    )
    gcrs_coo2 = gcrs_coo.transform_to(loc_gcrs)
    # Now we are relative to the correct observer, do the transform to CIRS.
    crepr = gcrs_coo2.cartesian.transform(pmat)
    return cirs_frame.realize_frame(crepr)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, CIRS, GCRS)
def cirs_to_gcrs(cirs_coo, gcrs_frame):
    # Compute the pmatrix, and then multiply by its transpose,
    pmat = gcrs_to_cirs_mat(cirs_coo.obstime)
    newrepr = cirs_coo.cartesian.transform(matrix_transpose(pmat))
    # We now have a GCRS vector for the input location and obstime.
    # Turn it into a GCRS frame instance.
    loc_gcrs = get_location_gcrs(
        cirs_coo.location, cirs_coo.obstime, cirs_to_itrs_mat(cirs_coo.obstime), pmat
    )
    gcrs = loc_gcrs.realize_frame(newrepr)
    # Finally, do any needed offsets (no-op if same obstime and location)
    return gcrs.transform_to(gcrs_frame)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, CIRS, ITRS)
def cirs_to_itrs(cirs_coo, itrs_frame):
    # first get us to CIRS at the target obstime, and location (no-op if same)
    cirs_coo2 = cirs_coo.transform_to(
        CIRS(obstime=itrs_frame.obstime, location=itrs_frame.location)
    )

    # now get the pmatrix
    pmat = cirs_to_itrs_mat(itrs_frame.obstime)
    crepr = cirs_coo2.cartesian.transform(pmat)
    return itrs_frame.realize_frame(crepr)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, ITRS, CIRS)
def itrs_to_cirs(itrs_coo, cirs_frame):
    # compute the pmatrix, and then multiply by its transpose
    pmat = cirs_to_itrs_mat(itrs_coo.obstime)
    newrepr = itrs_coo.cartesian.transform(matrix_transpose(pmat))
    cirs = CIRS(newrepr, obstime=itrs_coo.obstime, location=itrs_coo.location)

    # now do any needed offsets (no-op if same obstime and location)
    return cirs.transform_to(cirs_frame)


# TODO: implement GCRS<->CIRS if there's call for it.  The thing that's awkward
# is that they both have obstimes, so an extra set of transformations are necessary.
# so unless there's a specific need for that, better to just have it go through the above
# two steps anyway


@frame_transform_graph.transform(
    FunctionTransformWithFiniteDifference, GCRS, PrecessedGeocentric
)
def gcrs_to_precessedgeo(from_coo, to_frame):
    # first get us to GCRS with the right attributes (might be a no-op)
    gcrs_coo = from_coo.transform_to(
        GCRS(
            obstime=to_frame.obstime,
            obsgeoloc=to_frame.obsgeoloc,
            obsgeovel=to_frame.obsgeovel,
        )
    )

    # now precess to the requested equinox
    pmat = gcrs_precession_mat(to_frame.equinox)
    crepr = gcrs_coo.cartesian.transform(pmat)
    return to_frame.realize_frame(crepr)


@frame_transform_graph.transform(
    FunctionTransformWithFiniteDifference, PrecessedGeocentric, GCRS
)
def precessedgeo_to_gcrs(from_coo, to_frame):
    # first un-precess
    pmat = gcrs_precession_mat(from_coo.equinox)
    crepr = from_coo.cartesian.transform(matrix_transpose(pmat))
    gcrs_coo = GCRS(
        crepr,
        obstime=from_coo.obstime,
        obsgeoloc=from_coo.obsgeoloc,
        obsgeovel=from_coo.obsgeovel,
    )

    # then move to the GCRS that's actually desired
    return gcrs_coo.transform_to(to_frame)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, TEME, ITRS)
def teme_to_itrs(teme_coo, itrs_frame):
    # use the pmatrix to transform to ITRS in the source obstime
    pmat = teme_to_itrs_mat(teme_coo.obstime)
    crepr = teme_coo.cartesian.transform(pmat)
    itrs = ITRS(crepr, obstime=teme_coo.obstime)

    # transform the ITRS coordinate to the target obstime
    return itrs.transform_to(itrs_frame)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, ITRS, TEME)
def itrs_to_teme(itrs_coo, teme_frame):
    # transform the ITRS coordinate to the target obstime
    itrs_coo2 = itrs_coo.transform_to(ITRS(obstime=teme_frame.obstime))

    # compute the pmatrix, and then multiply by its transpose
    pmat = teme_to_itrs_mat(teme_frame.obstime)
    newrepr = itrs_coo2.cartesian.transform(matrix_transpose(pmat))
    return teme_frame.realize_frame(newrepr)


# Create loopback transformations
frame_transform_graph._add_merged_transform(ITRS, CIRS, ITRS)
frame_transform_graph._add_merged_transform(
    PrecessedGeocentric, GCRS, PrecessedGeocentric
)
frame_transform_graph._add_merged_transform(TEME, ITRS, TEME)
frame_transform_graph._add_merged_transform(TETE, ICRS, TETE)



# cython: language_level=3

cimport numpy as np
from libcpp cimport bool

np.import_array()


cdef extern from "src/convolve.h":
    void convolveNd_c(np.float64_t * const result,
            const np.float64_t * const f,
            const unsigned n_dim,
            const size_t * const image_shape,
            const np.float64_t * const g,
            const size_t * const kernel_shape,
            const bool nan_interpolate,
            const bool embed_result_within_padded_region,
            const unsigned n_threads) nogil


def _convolveNd_c(np.ndarray result,
                  np.ndarray array_to_convolve,
                  np.ndarray kernel,
                  bool nan_interpolate,
                  bool embed_result_within_padded_region,
                  int n_threads):
    convolveNd_c(
        <np.float64_t*>np.PyArray_DATA(result),
        <np.float64_t*>np.PyArray_DATA(array_to_convolve),
        array_to_convolve.ndim,
        <size_t*>array_to_convolve.shape,
        <np.float64_t*>np.PyArray_DATA(kernel),
        <size_t*>kernel.shape,
        nan_interpolate,
        embed_result_within_padded_region,
        n_threads,
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Astropy is a package intended to contain core functionality and some
common tools needed for performing astronomy and astrophysics research with
Python. It also provides an index for other astronomy packages and tools for
managing them.
"""

import sys
import warnings
from pathlib import Path

from astropy.utils.system_info import system_info

__all__ = [  # noqa: RUF100, RUF022
    "__version__",
    "__bibtex__",
    # Subpackages (mostly lazy-loaded)
    "config",
    "constants",
    "convolution",
    "coordinates",
    "cosmology",
    "io",
    "modeling",
    "nddata",
    "samp",
    "stats",
    "table",
    "tests",
    "time",
    "timeseries",
    "uncertainty",
    "units",
    "utils",
    "visualization",
    "wcs",
    # Functions
    "test",
    "log",
    "find_api_page",
    "online_help",
    "online_docs_root",
    "conf",
    "physical_constants",
    "astronomical_constants",
    "system_info",
]


def __getattr__(attr):
    if attr in __all__:
        from importlib import import_module

        return import_module("astropy." + attr)

    raise AttributeError(f"module 'astropy' has no attribute {attr!r}")


def __dir__():
    return sorted(set(globals()).union(__all__))


from . import config as _config
from .version import version as __version__


class Conf(_config.ConfigNamespace):
    """
    Configuration parameters for `astropy`.
    """

    unicode_output = _config.ConfigItem(
        False,
        "When True, use Unicode characters when outputting values, and "
        "displaying widgets at the console.",
    )
    use_color = _config.ConfigItem(
        sys.platform != "win32",
        "When True, use ANSI color escape sequences when writing to the console.",
        aliases=["astropy.utils.console.USE_COLOR", "astropy.logger.USE_COLOR"],
    )
    max_lines = _config.ConfigItem(
        None,
        description=(
            "Maximum number of lines in the display of pretty-printed "
            "objects. If not provided, try to determine automatically from the "
            "terminal size.  Negative numbers mean no limit."
        ),
        cfgtype="integer(default=None)",
        aliases=["astropy.table.pprint.max_lines"],
    )
    max_width = _config.ConfigItem(
        None,
        description=(
            "Maximum number of characters per line in the display of "
            "pretty-printed objects.  If not provided, try to determine "
            "automatically from the terminal size. Negative numbers mean no "
            "limit."
        ),
        cfgtype="integer(default=None)",
        aliases=["astropy.table.pprint.max_width"],
    )


conf = Conf()


# Define a base ScienceState for configuring constants and units
from .utils.state import ScienceState


class base_constants_version(ScienceState):
    """
    Base class for the real version-setters below.
    """

    _value = "test"

    _versions = dict(test="test")

    @classmethod
    def validate(cls, value):
        if value not in cls._versions:
            raise ValueError(f"Must be one of {list(cls._versions.keys())}")
        return cls._versions[value]

    @classmethod
    def set(cls, value):
        """
        Set the current constants value.
        """
        import sys

        if "astropy.units" in sys.modules:
            raise RuntimeError("astropy.units is already imported")
        if "astropy.constants" in sys.modules:
            raise RuntimeError("astropy.constants is already imported")

        return super().set(value)


class physical_constants(base_constants_version):
    """
    The version of physical constants to use.
    """

    # Maintainers: update when new constants are added
    _value = "codata2018"

    _versions = dict(
        codata2022="codata2022",
        codata2018="codata2018",
        codata2014="codata2014",
        codata2010="codata2010",
        astropyconst40="codata2018",
        astropyconst20="codata2014",
        astropyconst13="codata2010",
    )


class astronomical_constants(base_constants_version):
    """
    The version of astronomical constants to use.
    """

    # Maintainers: update when new constants are added
    _value = "iau2015"

    _versions = dict(
        iau2015="iau2015",
        iau2012="iau2012",
        astropyconst40="iau2015",
        astropyconst20="iau2015",
        astropyconst13="iau2012",
    )


# Create the test() function
from .tests.runner import TestRunner

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=PendingDeprecationWarning)
    test = TestRunner.make_test_runner_in(__path__[0])


# Set the bibtex entry to the article referenced in CITATION.
def _get_bibtex():
    refs = (Path(__file__).parent / "CITATION").read_text().split("@ARTICLE")[1:]
    return f"@ARTICLE{refs[0]}" if refs else ""


__citation__ = __bibtex__ = _get_bibtex()

from .logger import _init_log, _teardown_log

log = _init_log()

from types import ModuleType as __module_type__

from .utils.misc import find_api_page, online_docs_root, online_help

# Clean up top-level namespace--delete everything that isn't in __all__
# or is a magic attribute, and that isn't a submodule of this package
for varname in dir():
    if not (
        (varname.startswith("__") and varname.endswith("__"))
        or varname in __all__
        or (
            varname[0] != "_"
            and isinstance(locals()[varname], __module_type__)
            and locals()[varname].__name__.startswith(__name__ + ".")
        )
    ):
        # The last clause in the above disjunction deserves explanation:
        # When using relative imports like ``from .. import config``, the
        # ``config`` variable is automatically created in the namespace of
        # whatever module ``..`` resolves to (in this case astropy).  This
        # happens a few times just in the module setup above.  This allows
        # the cleanup to keep any public submodules of the astropy package
        del locals()[varname]

del varname, __module_type__



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# This module implements the base NDData class.


from copy import deepcopy

import numpy as np

from astropy import log
from astropy.units import Quantity, Unit
from astropy.utils.compat.optional_deps import HAS_DASK
from astropy.utils.masked import Masked, MaskedNDArray
from astropy.utils.metadata import MetaData
from astropy.wcs.wcsapi import (
    BaseHighLevelWCS,
    BaseLowLevelWCS,
    HighLevelWCSWrapper,
    SlicedLowLevelWCS,  # noqa: F401
)

from .nddata_base import NDDataBase
from .nduncertainty import NDUncertainty, UnknownUncertainty

__all__ = ["NDData"]

_meta_doc = """`dict`-like : Additional meta information about the dataset."""


class NDData(NDDataBase):
    """
    A container for `numpy.ndarray`-based datasets, using the
    `~astropy.nddata.NDDataBase` interface.

    The key distinction from raw `numpy.ndarray` is the presence of
    additional metadata such as uncertainty, mask, unit, a coordinate system
    and/or a dictionary containing further meta information. This class *only*
    provides a container for *storing* such datasets. For further functionality
    take a look at the ``See also`` section.

    See also: https://docs.astropy.org/en/stable/nddata/

    Parameters
    ----------
    data : `numpy.ndarray`-like or `NDData`-like
        The dataset.

    uncertainty : any type, optional
        Uncertainty in the dataset.
        Should have an attribute ``uncertainty_type`` that defines what kind of
        uncertainty is stored, for example ``"std"`` for standard deviation or
        ``"var"`` for variance. A metaclass defining such an interface is
        `NDUncertainty` - but isn't mandatory. If the uncertainty has no such
        attribute the uncertainty is stored as `UnknownUncertainty`.
        Defaults to ``None``.

    mask : any type, optional
        Mask for the dataset. Masks should follow the ``numpy`` convention that
        **valid** data points are marked by ``False`` and **invalid** ones with
        ``True``.
        Defaults to ``None``.

    wcs : any type, optional
        World coordinate system (WCS) for the dataset.
        Default is ``None``.

    meta : `dict`-like object, optional
        Additional meta information about the dataset. If no meta is provided
        an empty dict is created.
        Default is ``None``.

    unit : unit-like, optional
        Unit for the dataset. Strings that can be converted to a
        `~astropy.units.Unit` are allowed.
        Default is ``None``.

    copy : `bool`, optional
        Indicates whether to save the arguments as copy. ``True`` copies
        every attribute before saving it while ``False`` tries to save every
        parameter as reference.
        Note however that it is not always possible to save the input as
        reference.
        Default is ``False``.

        .. versionadded:: 1.2

    psf : `numpy.ndarray` or None, optional
        Image representation of the PSF. In order for convolution to be flux-
        preserving, this should generally be normalized to sum to unity.

    Raises
    ------
    TypeError
        In case ``data`` or ``meta`` don't meet the restrictions.

    Notes
    -----
    Each attribute can be accessed through the homonymous instance attribute:
    ``data`` in a `NDData` object can be accessed through the `data`
    attribute::

        >>> from astropy.nddata import NDData
        >>> nd = NDData([1,2,3])
        >>> nd.data
        array([1, 2, 3])

    Given a conflicting implicit and an explicit parameter during
    initialization, for example the ``data`` is a `~astropy.units.Quantity` and
    the unit parameter is not ``None``, then the implicit parameter is replaced
    (without conversion) by the explicit one and a warning is issued::

        >>> import numpy as np
        >>> import astropy.units as u
        >>> q = np.array([1,2,3,4]) * u.m
        >>> nd2 = NDData(q, unit=u.cm)
        INFO: overwriting Quantity's current unit with specified unit. [astropy.nddata.nddata]
        >>> nd2.data  # doctest: +FLOAT_CMP
        array([100., 200., 300., 400.])
        >>> nd2.unit
        Unit("cm")

    See Also
    --------
    NDDataRef
    NDDataArray
    """

    # Instead of a custom property use the MetaData descriptor also used for
    # Tables. It will check if the meta is dict-like or raise an exception.
    meta = MetaData(doc=_meta_doc, copy=False)

    def __init__(
        self,
        data,
        uncertainty=None,
        mask=None,
        wcs=None,
        meta=None,
        unit=None,
        copy=False,
        psf=None,
    ):
        # Rather pointless since the NDDataBase does not implement any setting
        # but before the NDDataBase did call the uncertainty
        # setter. But if anyone wants to alter this behavior again the call
        # to the superclass NDDataBase should be in here.
        super().__init__()

        # Check if data is any type from which to collect some implicitly
        # passed parameters.
        if isinstance(data, NDData):  # don't use self.__class__ (issue #4137)
            # Of course we need to check the data because subclasses with other
            # init-logic might be passed in here. We could skip these
            # tests if we compared for self.__class__ but that has other
            # drawbacks.

            # Comparing if there is an explicit and an implicit unit parameter.
            # If that is the case use the explicit one and issue a warning
            # that there might be a conflict. In case there is no explicit
            # unit just overwrite the unit parameter with the NDData.unit
            # and proceed as if that one was given as parameter. Same for the
            # other parameters.
            if unit is None and data.unit is not None:
                unit = data.unit
            elif unit is not None and data.unit is not None:
                log.info("overwriting NDData's current unit with specified unit.")

            if uncertainty is not None and data.uncertainty is not None:
                log.info(
                    "overwriting NDData's current "
                    "uncertainty with specified uncertainty."
                )
            elif data.uncertainty is not None:
                uncertainty = data.uncertainty

            if mask is not None and data.mask is not None:
                log.info("overwriting NDData's current mask with specified mask.")
            elif data.mask is not None:
                mask = data.mask

            if wcs is not None and data.wcs is not None:
                log.info("overwriting NDData's current wcs with specified wcs.")
            elif data.wcs is not None:
                wcs = data.wcs

            if psf is not None and data.psf is not None:
                log.info("Overwriting NDData's current psf with specified psf.")
            elif data.psf is not None:
                psf = data.psf

            if meta is not None and data.meta is not None:
                log.info("overwriting NDData's current meta with specified meta.")
            elif data.meta is not None:
                meta = data.meta

            # get the data attribute as it is, and continue to process it:
            data = data.data

        # if the data is wrapped by astropy.utils.masked.Masked:
        if isinstance(data, Masked):
            # first get the mask if one is available:
            if hasattr(data, "mask"):
                if mask is not None:
                    log.info(
                        "overwriting Masked Quantity's current mask with specified mask."
                    )
                else:
                    mask = data.mask

            if isinstance(data, MaskedNDArray):
                if unit is not None and hasattr(data, "unit") and data.unit != unit:
                    log.info(
                        "overwriting MaskedNDArray's current unit with specified unit."
                    )
                    data = data.to(unit).value
                elif unit is None and hasattr(data, "unit"):
                    unit = data.unit
                    data = data.value

                # now get the unmasked ndarray:
                data = np.asarray(data)

            if isinstance(data, Quantity):
                # this is a Quantity:
                if unit is not None and data.unit != unit:
                    log.info("overwriting Quantity's current unit with specified unit.")
                    data = data.to(unit)
                elif unit is None and data.unit is not None:
                    unit = data.unit
                data = data.value

        if isinstance(data, np.ma.masked_array):
            if mask is not None:
                log.info(
                    "overwriting masked ndarray's current mask with specified mask."
                )
            else:
                mask = data.mask
            data = data.data

        if isinstance(data, Quantity):
            # this is a Quantity:
            if unit is not None and data.unit != unit:
                log.info("overwriting Quantity's current unit with specified unit.")
                data = data.to(unit)
            elif unit is None and data.unit is not None:
                unit = data.unit
            data = data.value

        if isinstance(data, np.ndarray):
            # check for mask from np.ma.masked_ndarray
            if hasattr(data, "mask"):
                if mask is not None:
                    log.info(
                        "overwriting masked ndarray's current mask with specified mask."
                    )
                else:
                    mask = data.mask

        # Quick check on the parameters if they match the requirements.
        if (
            not hasattr(data, "shape")
            or not hasattr(data, "__getitem__")
            or not hasattr(data, "__array__")
        ):
            # Data doesn't look like a numpy array, try converting it to
            # one.
            data = np.asanyarray(data)
        # Another quick check to see if what we got looks like an array
        # rather than an object (since numpy will convert a
        # non-numerical/non-string inputs to an array of objects).
        if data.dtype == "O":
            raise TypeError("could not convert data to numpy array.")

        if unit is not None:
            unit = Unit(unit)

        if copy:
            # Data might have been copied before but no way of validating
            # without another variable.
            data = deepcopy(data)
            mask = deepcopy(mask)
            wcs = deepcopy(wcs)
            psf = deepcopy(psf)
            meta = deepcopy(meta)
            uncertainty = deepcopy(uncertainty)
            # Actually - copying the unit is unnecessary but better safe
            # than sorry :-)
            unit = deepcopy(unit)

        # Store the attributes
        self._data = data
        self.mask = mask
        self._wcs = None
        if wcs is not None:
            # Validate the wcs
            self.wcs = wcs
        self.meta = meta  # TODO: Make this call the setter sometime
        self._unit = unit
        # Call the setter for uncertainty to further check the uncertainty
        self.uncertainty = uncertainty
        self.psf = psf

    def __str__(self):
        data = str(self.data)
        unit = f" {self.unit}" if self.unit is not None else ""

        return data + unit

    def __repr__(self):
        prefix = self.__class__.__name__ + "("
        # to support reprs for other non-ndarray `data` attributes,
        # add more cases here:
        if HAS_DASK:
            import dask.array as da

            is_dask = isinstance(self.data, da.Array)
        else:
            is_dask = False

        if (
            isinstance(self.data, (int, float, np.ndarray))
            or np.issubdtype(float, self.data)
            or np.issubdtype(int, self.data)
        ) and not is_dask:
            # if data is an ndarray, get build a repr via Masked:
            ma = Masked(self.data, mask=self.mask)
            data_repr = repr(ma)
            # strip the class name generated by Masked, and the final ")",
            # and re-pad multi-line array outputs
            after_first_paren = data_repr.index("(") + 1
            old_prefix_spaces = " " * after_first_paren
            new_prefix_spaces = " " * len(prefix)
            data_repr = data_repr[after_first_paren:-1].replace(
                old_prefix_spaces, new_prefix_spaces
            )
            unit = f", unit='{self.unit}'" if self.unit is not None else ""
            return f"{prefix}{data_repr}{unit})"
        else:
            # relying on the builtin repr for objects other than ndarrays
            # allows for compatibility with e.g. dask:
            contents = []
            for attr in ("data", "mask", "uncertainty", "unit"):
                attr_data = getattr(self, attr)
                if attr_data is not None:
                    attr_prefix = f"\n  {attr}="
                    attr_repr = repr(attr_data)
                    attr_repr = attr_repr.replace(
                        "\n", f"\n{' ' * (len(attr_prefix) - 1)}"
                    )
                    contents.append(attr_prefix + attr_repr)
            return prefix + ",".join(contents) + "\n)"

    @property
    def data(self):
        """
        `~numpy.ndarray`-like : The stored dataset.
        """
        return self._data

    @property
    def mask(self):
        """
        any type : Mask for the dataset, if any.

        Masks should follow the ``numpy`` convention that valid data points are
        marked by ``False`` and invalid ones with ``True``.
        """
        return self._mask

    @mask.setter
    def mask(self, value):
        self._mask = value

    @property
    def unit(self):
        """
        `~astropy.units.Unit` : Unit for the dataset, if any.
        """
        return self._unit

    @property
    def wcs(self):
        """
        any type : A world coordinate system (WCS) for the dataset, if any.
        """
        return self._wcs

    @wcs.setter
    def wcs(self, wcs):
        if self._wcs is not None and wcs is not None:
            raise ValueError(
                "You can only set the wcs attribute with a WCS if no WCS is present."
            )

        if wcs is None or isinstance(wcs, BaseHighLevelWCS):
            self._wcs = wcs
        elif isinstance(wcs, BaseLowLevelWCS):
            self._wcs = HighLevelWCSWrapper(wcs)
        else:
            raise TypeError(
                "The wcs argument must implement either the high or low level WCS API."
            )

    @property
    def psf(self):
        return self._psf

    @psf.setter
    def psf(self, value):
        self._psf = value

    @property
    def uncertainty(self):
        """
        any type : Uncertainty in the dataset, if any.

        Should have an attribute ``uncertainty_type`` that defines what kind of
        uncertainty is stored, such as ``'std'`` for standard deviation or
        ``'var'`` for variance. A metaclass defining such an interface is
        `~astropy.nddata.NDUncertainty` but isn't mandatory.
        """
        return self._uncertainty

    @uncertainty.setter
    def uncertainty(self, value):
        if value is not None:
            # There is one requirements on the uncertainty: That
            # it has an attribute 'uncertainty_type'.
            # If it does not match this requirement convert it to an unknown
            # uncertainty.
            if not hasattr(value, "uncertainty_type"):
                log.info("uncertainty should have attribute uncertainty_type.")
                value = UnknownUncertainty(value, copy=False)

            # If it is a subclass of NDUncertainty we must set the
            # parent_nddata attribute. (#4152)
            if isinstance(value, NDUncertainty):
                # In case the uncertainty already has a parent create a new
                # instance because we need to assume that we don't want to
                # steal the uncertainty from another NDData object
                if value._parent_nddata is not None:
                    value = value.__class__(value, copy=False)
                # Then link it to this NDData instance (internally this needs
                # to be saved as weakref but that's done by NDUncertainty
                # setter).
                value.parent_nddata = self
        self._uncertainty = value
